"""Data Buffer Wrapper
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

class _Dimension():
    def __init__(self, data_buffer, group_id, channel_id, dimension_id):
        self._data_buffer_com = data_buffer
        self._group_id = group_id
        self._channel_id = channel_id
        self._dimension_id = dimension_id

    @property
    def name(self):
        return self._data_buffer_com.DataName(
            self._group_id, self._channel_id, self._dimension_id)

    @property
    def unit(self):
        return self._data_buffer_com.DataUnit(
            self._group_id, self._channel_id, self._dimension_id)

    @property
    def data_min(self):
        return self._data_buffer_com.DataMin(
            self._group_id, self._channel_id, self._dimension_id)

    @property
    def data_range(self):
        return self._data_buffer_com.DataRange(
            self._group_id, self._channel_id, self._dimension_id)


class _Channel():
    def __init__(self, data_buffer, group_id, channel_id):
        # TODO: can we not access the properties of the containing class ?
        self._data_buffer_com = data_buffer
        self._group_id = group_id
        self._channel_id = channel_id

    @property
    def data(self):
        # If we get the data as soon as we are not syncing and the data is
        # valid, we nevertheless get back None.
        # If we try a little later, we get valid data.
        # TODO: why is this neccessary ?
        tmp_data = None
        while tmp_data is None:
            tmp_data = self._data_buffer_com.DataRAW(
                self._group_id, self._channel_id)
        return tmp_data

    # Transfer of raw data from PC to controller (e.g. for memory signal generator)
    #
    # The raw data is expected to be in channel interleaved format
    # with the channel_id set to the channel mask.
    # Memory signal generator has fixed 8 channel data format therefore
    # channel_id must be set to 255 and the data has the interleaved format
    # ch1[0] ch2[0] ch3[0] ch4[0] ch5[0] ch6[0] ch7[0] ch8[0] ch1[1] ch2[1], ch3[1], ...
    #
    # example:
    # group = ll.DataBuffer()
    # channel = group.channel(255) # channel mask always 8 interleaved channels!
    # d = np.array(d1)
    # buffer = np.empty((d.size * 8,))
    # buffer[0::8] = d1 # Position x
    # buffer[1::8] = d2 # Drive x
    # buffer[2::8] = d3 # Position y
    # buffer[3::8] = d4 # Drive y
    # buffer[4::8] = d5 # Position z
    # buffer[5::8] = d6 # Drive Z
    # buffer[6::8] = d7 # Out7
    # buffer[7::8] = d8 # Out8
    # channel.data = buffer

    @data.setter
    def data(self, my_data):
        self._data_buffer_com.DataRAW(
            self._group_id, self._channel_id, my_data)

    def filtered_data(self, filter_id):
        tmp_data = None
        while tmp_data is None:
            tmp_data = self._data_buffer_com.Data(
                self._group_id, self._channel_id, filter_id)
        return tmp_data

    def dimension(self, dimension_id):
        return _Dimension(
            self._data_buffer_com, self._group_id, self._channel_id,
            dimension_id)


class DataBufferInterface():
    """Python Interface to the DataBuffer COM object."""
    # Reference to a DataBuffer COM object,
    # Value should be added set on creation of sub-class by data_buffer_type.
    # Thus, each COM interface application interace has its own Python
    # interace.
    _data_buffer_com = None

    def __init__(self, group_id=-1):
        if group_id == -1:
            self._group_id = self._data_buffer_com.CreateDataGroup
            self._new_group = True
        else:
            self._group_id = group_id
            self._new_group = False

    def __del__(self):
        # TODO:
        # - make destructor instead
        # - it (or the spm object) should check if the DataGroup is still in
        #   use
        if self._new_group:
            self._data_buffer_com.DeleteDataGroup(self._group_id)

    def channel(self, channel_id: int):
        """Returns a channel of the data buffer.

        Parameters
        ----------
        channel_id:
            the index of the channel to create.
        """
        return _Channel(self._data_buffer_com, self._group_id, channel_id)

    def synchronize_data_group(self):
        self._data_buffer_com.SyncDataGroup(self._group_id)

    @classmethod
    def delete_all(cls):
        cls._data_buffer_com.DeleteAllDataGroups()

    @property
    def available_points(cls):
        return cls._data_buffer_com.AvailableBufferPoints

    @property
    def count(cls):
        return cls._data_buffer_com.DataGroupCount

    @property
    def group_id(self):
        return self._group_id

    @property
    def is_synchronizing(self):
        return self._data_buffer_com.IsSyncingDataGroup(self._group_id)

    @property
    def is_valid(self):
        return self._data_buffer_com.DataValid(self._group_id)

    @property
    def timestamp_first_sample(self):
        return self._data_buffer_com.TimestampFirstSample(self._group_id)

    @property
    def timestamp_first_sample_str(self):
        return self._data_buffer_com.TimestampFirstSampleStr(self._group_id)

    @property
    def timestamp_last_sample(self):
        return self._data_buffer_com.TimestampLastSample(self._group_id)

    @property
    def timestamp_last_sample_str(self):
        return self._data_buffer_com.TimestampLastSampleStr(self._group_id)

    @property
    def channel_count(self):
        return self._data_buffer_com.DataChannelCount(self._group_id)


def data_buffer_type(data_buffer):
    """Creates a Python databuffer interface type for a DataBuffer COM object.
    To be used by the LowLevel() class.

    Returns
    -------
    DataBufferInterface derived object type.

    Parameters
    ----------
    data_buffer: DataBuffer COM object.
    """
    type_dict = {'_data_buffer_com': data_buffer}
    return type('DataBuffer', (DataBufferInterface,), type_dict)



"""Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import enum
import nanosurf.lib.devices.i2c.bus_master as i2c 

class Chip_TCN75(i2c.I2CChip): 
    """ temperature sensor chip"""
    class Resolution(enum.IntEnum):
        Bit9 = 0
        Bit10 = 1
        Bit11 = 2
        Bit12 = 3

    def __init__(self, bus_addr: int, **kwargs):
        super().__init__(bus_addr, offset_mode=i2c.I2COffsetMode.U8Bit,**kwargs)

    @property
    def reg_config(self) -> int:
        return self.read_byte_with_offset(0x01)

    @reg_config.setter
    def reg_config(self, config: int):
        self.write_byte_with_offset(0x01, config)

    @property
    def reg_temp(self) -> int:
        val = self.read_bytes_with_offset(0x00,2)
        return val[0]*256+val[1]

    @property
    def reg_hysteresis(self) -> int:
        val = self.read_bytes_with_offset(0x02,2)
        return val[0]*256+val[1]
        
    @reg_hysteresis.setter
    def reg_hysteresis(self, val: int):
        val = self.write_bytes_with_offset(0x02,[val/256, val%256])
        
    @property
    def reg_limit(self) -> int:
        val = self.read_bytes_with_offset(0x03,2)
        return val[0]*256+val[1]
        
    @reg_limit.setter
    def reg_hysteresis(self, val: int):
        val = self.write_bytes_with_offset(0x03,[val/256, val%256])
        
    @property
    def temperature(self) -> float:
        """ Get current chip temperature in degree. Read only"""
        temp = 1.0 / 256.0 * self.reg_temp
        return temp 

    @property
    def hysteresis(self) -> float:
        temp = 1.0 / 256.0 * self.reg_hysteresis
        return temp 

    @hysteresis.setter
    def hysteresis(self, val: float):
        self.reg_hysteresis = int(val * 256.0)

    @property
    def temp_limit(self) -> float:
        temp = 1.0 / 256.0 * self.reg_limit
        return temp 

    @temp_limit.setter
    def temp_limit(self, val: float):
        self.reg_limit = int(val * 256.0)

    @property
    def resolution(self) -> Resolution:
        """ Defines bit resolution of temp measurement. 
        From 9Bit -> 0.5deg up to 12Bit -> 0.0625deg"""
        resolution_bits = int((self.reg_config >> 5) & 0x03)
        if resolution_bits == 0:
            return self.Resolution.Bit9
        elif resolution_bits == 1:
              return self.Resolution.Bit10
        elif resolution_bits == 2:
              return self.Resolution.Bit11
        elif resolution_bits == 3:
              return self.Resolution.Bit12
        raise NotImplementedError
            
    @resolution.setter
    def resolution(self, res: Resolution):
        config_value = self.reg_config & ~(0x03 << 5)
        self.reg_config = config_value | (res << 5)
       



"""Frequency Sweep configuration and execution
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import enum
import matplotlib.pyplot as plt
import numpy as np
import time

import nanosurf
import nanosurf.lib.spm.com_proxy as spm

class _ModulationOutput():
    def __init__(self):
        pass

    def __del__(self):
        pass


class _NormalExcitation(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._fast_out1 = lowlevel.AnalogFastOut(
            lowlevel.AnalogFastOut.Instance.EXCITATION)
        self._old_input = self._fast_out1.input.value
        self._fast_out1.input.value = self._fast_out1.InputChannels.Analyzer2_Reference

    def __del__(self):
        self._fast_out1.input.value = self._old_input
        super().__del__()


class _TipVoltage(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._fast_out2 = lowlevel.AnalogFastOut(
            lowlevel.AnalogFastOut.Instance.USER)
        self._hi_res_tip_voltage = lowlevel.AnalogHiResOut(
            lowlevel.AnalogHiResOut.Instance.TIPVOLTAGE)
        self._old_fast2_input = self._fast_out2.input.value
        self._old_tip_voltage_modulation = self._hi_res_tip_voltage.modulation.value
        self._fast_out2.input.value = self._fast_out2.InputChannels.Analyzer2_Reference
        self._hi_res_tip_voltage.modulation.value = (
                self._hi_res_tip_voltage.Modulation.Enabled)

    def __del__(self):
        self._fast_out2.input.value = self._old_fast2_input
        self._hi_res_tip_voltage.modulation.value = self._old_tip_voltage_modulation
        super().__del__()


class _FastUser(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._fast_out2 = lowlevel.AnalogFastOut(
            lowlevel.AnalogFastOut.Instance.USER)
        self._old_fast2_input = self._fast_out2.input.value
        self._old_fast2_analog = self._fast_out2.analog_output.value
        self._fast_out2.input.value = self._fast_out2.InputChannels.Analyzer2_Reference
        self._fast_out2.analog_output.value = (
                self._fast_out2.AnalogOutput.Enabled)

    def __del__(self):
        self._fast_out2.input.value = self._old_fast2_input
        self._fast_out2.analog_output.value = self._old_fast2_analog
        super().__del__()


class _OutUser1(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._fast_out2 = lowlevel.AnalogFastOut(
            lowlevel.AnalogFastOut.Instance.USER)
        self._hi_res_user_1 = lowlevel.AnalogHiResOut(
            lowlevel.AnalogHiResOut.Instance.USER1)
        self._old_fast2_input = self._fast_out2.input.value
        self._old_hires4_modulation = self._hi_res_user_1.modulation.value
        self._fast_out2.input.value = self._fast_out2.InputChannels.Analyzer2_Reference
        self._hi_res_user_1.modulation.value = (
                self._hi_res_user_1.Modulation.Enabled)

    def __del__(self):
        self._fast_out2.input.value = self._old_fast2_input
        self._hi_res_user_1.modulation.value = self._old_hires4_modulation
        super().__del__()


class _OutUser2(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._fast_out2 = lowlevel.AnalogFastOut(
            lowlevel.AnalogFastOut.Instance.USER)
        self._hi_res_user2 = lowlevel.AnalogHiResOut(
            lowlevel.AnalogHiResOut.Instance.USER2)
        self._old_fast2_input = self._fast_out2.input.value
        self._old_user2_modulation = self._hi_res_user2.modulation.value
        self._fast_out2.input.value = self._fast_out2.InputChannels.Analyzer2_Reference
        self._hi_res_user2.modulation.value = (
                self._hi_res_user2.Modulation.Enabled)

    def __del__(self):
        self._fast_out2.input.value = self._old_fast2_input
        self._hi_res_user2.modulation.value = self._old_user2_modulation
        super().__del__()


class _OutUser3(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._fast_out2 = lowlevel.AnalogFastOut(
            lowlevel.AnalogFastOut.Instance.USER)
        self._out_user3 = lowlevel.AnalogHiResOut(
            lowlevel.AnalogHiResOut.Instance.USER3)
        self._old_fast2_input = self._fast_out2.input.value
        self._old_user3_modulation = self._out_user3.modulation.value
        self._fast_out2.input.value = self._fast_out2.InputChannels.Analyzer2_Reference
        self._out_user3.modulation.value = (
                self._out_user3.Modulation.Enabled)

    def __del__(self):
        self._fast_out2.input.value = self._old_fast2_input
        self._out_user3.modulation.value = self._old_user3_modulation
        super().__del__()


class _OutUser4(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._fast_out2 = lowlevel.AnalogFastOut(
            lowlevel.AnalogFastOut.Instance.USER)
        self._monitor_out2 = lowlevel.AnalogHiResOut(
            lowlevel.AnalogHiResOut.Instance.USER4)
        self._old_fast2_input = self._fast_out2.input.value
        self._old_monitor2_modulation = self._monitor_out2.modulation.value
        self._fast_out2.input.value = self._fast_out2.InputChannels.Analyzer2_Reference
        self._monitor_out2.modulation.value = (
                self._monitor_out2.Modulation.Enabled)

    def __del__(self):
        self._fast_out2.input.value = self._old_fast2_input
        self._monitor_out2.modulation.value = self._old_monitor2_modulation
        super().__del__()


class _OutPositionX(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._scan_x_out = lowlevel.AnalogHiResOut(
                lowlevel.AnalogHiResOut.Instance.POSITIONX)
        self._old_scan_x_out = self._scan_x_out.input.value
        self._scan_x_out.input.value = self._scan_x_out.InputChannels.Analyzer2_Reference

    def __del__(self):
        self._scan_x_out.input.value = self._old_scan_x_out
        super().__del__()


class _OutPositionY(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._scan_y_out = lowlevel.AnalogHiResOut(
                lowlevel.AnalogHiResOut.Instance.POSITIONY)
        self._old_scan_y_out = self._scan_y_out.input.value
        self._scan_y_out.input.value = self._scan_y_out.InputChannels.Analyzer2_Reference

    def __del__(self):
        self._scan_y_out.input.value = self._old_scan_y_out
        super().__del__()


class _OutPositionZ(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._hi_res_pos_out_z = lowlevel.AnalogHiResOut(
            lowlevel.AnalogHiResOut.Instance.POSITIONZ)
        self._old_posz_input = self._hi_res_pos_out_z.input.value
        self._hi_res_pos_out_z.input.value = (
                self._hi_res_pos_out_z.InputChannels.Analyzer2_Reference)

    def __del__(self):
        self._hi_res_pos_out_z.input.value = self._old_posz_input
        super().__del__()


class _ModOutZ(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._hi_res_pos_out_z = lowlevel.AnalogHiResOut(
            lowlevel.AnalogHiResOut.Instance.POSITIONZ)
        self._old_posz_modulation = self._hi_res_pos_out_z.modulation.value
        self._hi_res_pos_out_z.modulation.value = (
                self._hi_res_pos_out_z.Modulation.Enabled)

    def __del__(self):
        self._hi_res_pos_out_z.modulation.value = self._old_posz_modulation
        super().__del__()


class _ModXControlSet(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._pid_control = lowlevel.PIDController(lowlevel.PIDController.Instance.PIDX)
        try:
            self._old_value = self._pid_control.set_point_modulation_enable.value
            self._pid_control.set_point_modulation_enable.value = (
                self._pid_control.Enable.Enabled)
            self.new_style = True
        except:
            self._old_value = self._pid_control.select_sweep.value
            self._pid_control.select_sweep.value = (
                self._pid_control.SelectSweep.Selected)
            self.new_style = False


    def __del__(self):
        if self.new_style:
            self._pid_control.set_point_modulation_enable.value = self._old_value
        else:
            self._pid_control.select_sweep.value = self._old_value
        super().__del__()


class _ModYControlSet(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._pid_control = lowlevel.PIDController(lowlevel.PIDController.Instance.PIDY)
        try:
            self._old_value = self._pid_control.set_point_modulation_enable.value
            self._pid_control.set_point_modulation_enable.value = (
                self._pid_control.Enable.Enabled)
            self.new_style = True
        except:
            self._old_value = self._pid_control.select_sweep.value
            self._pid_control.select_sweep.value = (
                self._pid_control.SelectSweep.Selected)
            self.new_style = False

    def __del__(self):
        if self.new_style:
            self._pid_control.set_point_modulation_enable.value = self._old_value
        else:
            self._pid_control.select_sweep.value = self._old_value
        super().__del__()


class _ModZControlSet(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._pid_control = lowlevel.ZControllerEx()
        try:
            self._old_value = self._pid_control.set_point_modulation_enable.value
            self._pid_control.set_point_modulation_enable.value = (
                self._pid_control.Enable.Enabled)
            self.new_style = True
        except:
            self._old_value = self._pid_control.select_sweep.value
            self._pid_control.select_sweep.value = (
                self._pid_control.SelectSweep.Selected)
            self.new_style = False

    def __del__(self):
        if self.new_style:
            self._pid_control.set_point_modulation_enable.value = self._old_value
        else:
            self._pid_control.select_sweep.value = self._old_value
        super().__del__()


class _ModZControlOutput(_ModulationOutput):
    def __init__(self, lowlevel):
        super().__init__()
        self._pid_control = lowlevel.ZControllerEx()
        try:
            self._old_value = self._pid_control.output_modulation_enable.value
            self._pid_control.output_modulation_enable.value = (
                self._pid_control.Enable.Enabled)
            self.new_style = True
        except:
            self._old_value = self._pid_control.select_sweep.value
            self._pid_control.select_sweep.value = (
                self._pid_control.SelectSweep.Selected)
            self.new_style = False

    def __del__(self):
        if self.new_style:
            self._pid_control.output_modulation_enable.value = self._old_value
        else:
            self._pid_control.select_sweep.value = self._old_value
        super().__del__()

class FrequencySweepOutput(enum.Enum):
    """Enumeration for the modulation output selection"""
    Normal_Excitation = enum.auto(),
    FastUser_OutB = enum.auto(),
    TipVoltage_UserOutA = enum.auto(),
    UserOut1_UserC = enum.auto(),
    UserOut2_UserA = enum.auto(),
    UserOut3_Monitor1 = enum.auto(),
    UserOut4_Monitor2 = enum.auto(),
    PositionX = enum.auto(),
    PositionY = enum.auto(),
    PositionZ = enum.auto(),
    ModOutZ = enum.auto(),
    ModXControlSet = enum.auto(),
    ModYControlSet = enum.auto(),
    ModZControlSet = enum.auto(),
    ModZControlOutput = enum.auto()

class InputSource(enum.Enum):
    Deflection = enum.auto(),
    Fast2_CX = enum.auto(),
    Fast_User_CX = enum.auto(),
    Friction = enum.auto(),
    UserIn1 = enum.auto(),
    UserIn2_UserB = enum.auto(),
    UserIn3_UserA = enum.auto(),
    UserIn4_CX = enum.auto(),
    TipCurrent = enum.auto(),
    TestGND_C3000 = enum.auto(),
    TesRef_C3000 = enum.auto(),
    TestMixedOut3_C3000 = enum.auto(),
    AxisInX = enum.auto(),
    AxisInY = enum.auto(),
    AxisInZ = enum.auto(),
    MainLockIn_Amplitude = enum.auto(),
    MainPLL_FreqShift = enum.auto(),
    ZControllerOut = enum.auto()

class InputRanges(enum.Enum):
    Full = enum.auto(),
    OneOverFour = enum.auto(),
    OneOverSixteen = enum.auto(),

class Bandwidths(enum.IntEnum):
    Hz_23 = 0,
    Hz_45 = 1,
    Hz_90 = 2,
    Hz_180 = 3,
    Hz_360 = 4,
    Hz_740 = 5,
    Hz_1500 = 6,
    Hz_3000 = 7,
    Hz_6000 = 8,
    Hz_12k = 9,
    Hz_24k= 10,
    Hz_48k = 11,
    Hz_100k = 12,
    Hz_230k = 13,
    Hz_500k = 14

class FrequencySweep():
    """Workflow for acquiring and plotting frequency sweeps."""

    # Output = enum.Enum('Output', {
    Output = {
        FrequencySweepOutput.Normal_Excitation.name: _NormalExcitation,
        FrequencySweepOutput.FastUser_OutB.name: _FastUser,
        FrequencySweepOutput.TipVoltage_UserOutA.name: _TipVoltage,
        FrequencySweepOutput.UserOut1_UserC.name: _OutUser1,
        FrequencySweepOutput.UserOut2_UserA.name: _OutUser2,
        FrequencySweepOutput.UserOut3_Monitor1.name: _OutUser3,
        FrequencySweepOutput.UserOut4_Monitor2.name: _OutUser4,
        FrequencySweepOutput.PositionX.name: _OutPositionX,
        FrequencySweepOutput.PositionY.name: _OutPositionY,
        FrequencySweepOutput.PositionZ.name: _OutPositionZ,
        FrequencySweepOutput.ModOutZ.name: _ModOutZ,
        FrequencySweepOutput.ModXControlSet.name: _ModXControlSet,
        FrequencySweepOutput.ModYControlSet.name: _ModYControlSet,
        FrequencySweepOutput.ModZControlSet.name: _ModZControlSet,
        FrequencySweepOutput.ModZControlOutput.name: _ModZControlOutput
        }

    input_sources_names: dict[InputSource, str] = {
        InputSource.Deflection : "Deflection",
        InputSource.Fast2_CX :  "Fast2 (CX)",
        InputSource.Fast_User_CX : "Fast User (CX)",
        InputSource.Friction : "Friction",
        InputSource.UserIn1: "User In 1",
        InputSource.UserIn2_UserB : "User In 2 / B",
        InputSource.UserIn3_UserA : "User In 3 / A",
        InputSource.UserIn4_CX : "User In 4 (CX)",
        InputSource.TipCurrent : "TipCurrent",
        InputSource.TestGND_C3000 : "Test GND (C3000)",
        InputSource.TesRef_C3000 : "Test Ref (C3000)",
        InputSource.TestMixedOut3_C3000 : "Test MixedOut3 (C3000)",
        InputSource.AxisInX : "Axis In X",
        InputSource.AxisInY : "Axis In Y",
        InputSource.AxisInZ : "Axis In Z",
        InputSource.MainLockIn_Amplitude : "Main Lock-In Amplitude",
        InputSource.MainPLL_FreqShift : "Main PLL Frequency Shift",
        InputSource.ZControllerOut : "Z-Controller Out"
    }

    bandwidths_names = {
        Bandwidths.Hz_23: "23Hz",
        Bandwidths.Hz_45:"45Hz",
        Bandwidths.Hz_90:"90Hz",
        Bandwidths.Hz_180:"180Hz",
        Bandwidths.Hz_360:"360Hz",
        Bandwidths.Hz_740:"740Hz",
        Bandwidths.Hz_1500:"1500Hz",
        Bandwidths.Hz_3000:"3kHz",
        Bandwidths.Hz_6000:"6kHz",
        Bandwidths.Hz_12k:"12kHz",
        Bandwidths.Hz_24k:"23kHz",
        Bandwidths.Hz_48k:"45kHz",
        Bandwidths.Hz_100k:"100kHz",
        Bandwidths.Hz_230k:"230kHz",
        Bandwidths.Hz_500k:"500kHz"
    }

    input_ranges_names: dict[InputRanges,str] = {
        InputRanges.Full:           "Full",
        InputRanges.OneOverFour:    "1/4",
        InputRanges.OneOverSixteen: "1/16"
    }

    output_names: dict[FrequencySweepOutput, str] = {
        FrequencySweepOutput.Normal_Excitation:"Normal Excitation",
        FrequencySweepOutput.FastUser_OutB: "Fast User / Out B",
        FrequencySweepOutput.TipVoltage_UserOutA:"Tip Voltage / User Out A",
        FrequencySweepOutput.UserOut1_UserC:"User 1 / C",
        FrequencySweepOutput.UserOut2_UserA:"User 2 / A",
        FrequencySweepOutput.UserOut3_Monitor1:"User 3 / Monitor 1",
        FrequencySweepOutput.UserOut4_Monitor2:"User 4 / Monitor 2",
        FrequencySweepOutput.PositionX:"X Position",
        FrequencySweepOutput.PositionY:"Y Position",
        FrequencySweepOutput.PositionZ:"PosOutZ",
        FrequencySweepOutput.ModOutZ:"ModOutZ",
        FrequencySweepOutput.ModXControlSet:"ModXControlSet",
        FrequencySweepOutput.ModYControlSet:"ModYControlSet",
        FrequencySweepOutput.ModZControlSet:"ModZControlSet",
        FrequencySweepOutput.ModZControlOutput:"ModZControlOutput"
    }

    def __init__(self, spm: spm.Spm):
        self.is_busy = False
        self.output_setup = None
        self.data_group_name = "Frequency sweep"
        self.data_channel_amplitude = 2
        self.data_channel_phase = 3
        self._spm = spm
        self._ll = spm.lowlevel
        self._internal_transfer_time = 5.0

        self._analyzer = self._ll.SignalAnalyzer(spm.lowlevel.SignalAnalyzer.Instance.INST2)
        self._system_infra = self._ll.SystemInfra()
        self._frequency_sweep_generator = self._ll.FrequencySweepGen()

        self.input_sources_to_lu_map: dict[InputSource, int] = {
            InputSource.Deflection: self._analyzer.Input.FastInDeflection,
            InputSource.Fast2_CX: self._analyzer.Input.FastIn2,
            InputSource.Fast_User_CX: self._analyzer.Input.FastInUser,
            InputSource.Friction: self._analyzer.Input.InLateral,
            InputSource.UserIn1: self._analyzer.Input.InUser1,
            InputSource.UserIn2_UserB: self._analyzer.Input.InUser2,
            InputSource.UserIn3_UserA: self._analyzer.Input.InUser3,
            InputSource.UserIn4_CX: self._analyzer.Input.InUser4,
            InputSource.TipCurrent: self._analyzer.Input.InTipCurrent,
            InputSource.TestGND_C3000: self._analyzer.Input.Test_AnaGND,
            InputSource.TesRef_C3000: self._analyzer.Input.Test_Ref,
            InputSource.TestMixedOut3_C3000: self._analyzer.Input.Test_TipVoltage,
            InputSource.AxisInX: self._analyzer.Input.InPositionX,
            InputSource.AxisInY: self._analyzer.Input.InPositionY,
            InputSource.AxisInZ: self._analyzer.Input.InPositionZ,
            InputSource.MainLockIn_Amplitude: self._analyzer.Input.Analyzer1_Amplitude,
            InputSource.MainPLL_FreqShift: self._analyzer.Input.Analyzer1_CtrlDeltaF,
            InputSource.ZControllerOut: self._analyzer.Input.CtrlZ_Out,
        }

        self.input_range_to_gain_map: dict[InputRanges, float] = {
            InputRanges.Full:           self._system_infra.main_in_2_gain.value_min,
            InputRanges.OneOverFour:    self._system_infra.main_in_2_gain.value_max / self._system_infra.main_in_2_gain.value_max,
            InputRanges.OneOverSixteen: self._system_infra.main_in_2_gain.value_max
        }

        self.bandwidths_to_lu_map : dict[Bandwidths, int] = {
            Bandwidths.Hz_23 : self._analyzer.DemodulatorBW.BW_23Hz,
            Bandwidths.Hz_45: self._analyzer.DemodulatorBW.BW_45Hz,
            Bandwidths.Hz_90: self._analyzer.DemodulatorBW.BW_90Hz,
            Bandwidths.Hz_180: self._analyzer.DemodulatorBW.BW_180Hz,
            Bandwidths.Hz_360: self._analyzer.DemodulatorBW.BW_360Hz,
            Bandwidths.Hz_740: self._analyzer.DemodulatorBW.BW_750Hz,
            Bandwidths.Hz_1500: self._analyzer.DemodulatorBW.BW_1500Hz,
            Bandwidths.Hz_3000: self._analyzer.DemodulatorBW.BW_3kHz,
            Bandwidths.Hz_6000: self._analyzer.DemodulatorBW.BW_6kHz,
            Bandwidths.Hz_12k: self._analyzer.DemodulatorBW.BW_12kHz,
            Bandwidths.Hz_24k: self._analyzer.DemodulatorBW.BW_23kHz,
            Bandwidths.Hz_48k: self._analyzer.DemodulatorBW.BW_45kHz,
            Bandwidths.Hz_100k: self._analyzer.DemodulatorBW.BW_100kHz,
            Bandwidths.Hz_230k: self._analyzer.DemodulatorBW.BW_230kHz,
            Bandwidths.Hz_500k: self._analyzer.DemodulatorBW.BW_500kHz,
            }

    def __del__(self):
        if self.output_setup is not None:
            del self.output_setup

    def _get_active_data(self) -> tuple[np.ndarray, np.ndarray]:
        """Returns transfer function and frequencies of the active chart."""

        class ActiveError():
            """Exception raised when there is no active data."""
            def __init__(self, expression, message):
                self._expression = expression
                self._message = message

        def check_if_active(document):
            if document is None:
                raise ActiveError(document, "No document selected!")

        document = self._spm.application.DocGetActive
        check_if_active(document)

        group = document.DataGetGroupPos(self.data_group_name)

        new_data = document.DataGetByPos(group, self.data_channel_amplitude)
        assert new_data is not None, "No data!"
        data_line:str = new_data.GetLine(0, spm.Spm.DataFilter.RAW, spm.Spm.DataConversion.Physical)
        # in some countries numbers have ',' as dot but float(str) has problems with this localization
        usa_style_data_line = data_line.replace(",", ".")
        amplitude = np.array(usa_style_data_line.split(";")).astype(float)

        new_data = document.DataGetByPos(group, self.data_channel_phase)
        assert new_data is not None, "No data!"
        data_line:str = new_data.GetLine(0, spm.Spm.DataFilter.RAW, spm.Spm.DataConversion.Physical)
        # in some countries numbers have ',' as dot but float(str) has problems with this localisation
        usa_style_data_line = data_line.replace(",", ".")
        phase = np.array(usa_style_data_line.split(";")).astype(float)

        frequencies = np.arange(
            new_data.AxisPointMin,
            new_data.AxisPointMin + new_data.AxisPointRange + 0.1,
            new_data.AxisPointRange / (new_data.Points - 1))
        complex_transfer = amplitude * np.exp(1j * 2*np.pi * phase/360.0)

        return (complex_transfer, frequencies)

    def _start(
            self, start_frequency: float, end_frequency: float, points: int,
            step_time: float, settle_time: float, amplitude: float) -> float:
        """Starts Frequency Sweep.

        Parameters
        ----------
        start_frequency
            Start frequency [Hz]
        end_frequency
            End frequency [Hz]
        points
            Number of points in the sweep.
        step_time
            Time between measurement points. [s]
        settle_time
            Time to wait between setting up signal_analyzer and
            starting the frequency sweep. [s]
        amplitude: float
            Normalized Amplitude [0..1.0]
        """
        limited_step_time = max(step_time, 0.01)

        generator = self._frequency_sweep_generator
        generator.lusig_analyzer_inst_no.value = (
            generator.LUSigAnalyzerInstNo.INST2)

        generator.start_frequency.value = start_frequency
        generator.end_frequency.value = end_frequency
        generator.data_points.value = points

        # Settle time before starting the measurement...
        generator.settle_time.value = settle_time
        generator.step_time.value = limited_step_time
        generator.sweep_amplitude.value = amplitude

        self.start_time = time.time()
        self.start_frequency = start_frequency
        self.end_frequency = end_frequency
        self.is_busy = True
        generator.start_frequency_sweep()

        return limited_step_time*points + 2*settle_time + self._internal_transfer_time/2

    def _finish(self):
        """Stop FrequencySweep and capture measured data into new document"""
        self._frequency_sweep_generator.user_abort()
        time.sleep(0.1)
        self._spm.application.OperatingMode.CaptureFreqSearchChart
        time.sleep(self._internal_transfer_time/2)
        self.is_busy = False

    def _to_hz(self, mixer_bw_sel: Bandwidths) -> float:
        """Converts MixerBW enumeration to bandwidth in Hz."""
        mixer_base_filter_frequency = 11.1
        mixer_filter_step_factor = 2.0
        return mixer_base_filter_frequency * mixer_filter_step_factor**mixer_bw_sel.value

    def start_execute(
            self, start_frequency: float, end_frequency: float, frequency_step: float,
            sweep_amplitude: float,
            input_source: InputSource,
            input_range: InputRanges,
            mixer_bw_select: Bandwidths,
            reference_phase: float, output: FrequencySweepOutput) -> float:
        """Prepares and executes the Frequency Sweep.

        Parameters
        ----------
        start_frequency
            Start frequency [Hz]
        end_frequency
            End frequency [Hz]
        frequency_step
            Frequency difference between two consecutive points [Hz]
        sweep_amplitude
            Normalized Amplitude [0..1.0]
        input_source: InputSource
            Lock-in input channel
        input_range: InputRanges
            Input range switch
        mixer_bw_select : Bandwidths
            Mixer bandwidth
        reference_phase : float
            Phase shift of the reference output relative to the internal reference [°]
        output :  FrequencySweepOutput
            Selects the output

        Returns
        ----------
        total_time: float
            returns the calculated measuring time in [s]
        """
        sweep_amplitude = max(min(float(sweep_amplitude), 1.0), 0.0)
        step_time = 1/self._to_hz(mixer_bw_select)
        settle_time = 100 * step_time
        points = int((end_frequency - start_frequency)/frequency_step) + 1

        self._analyzer.operating_mode.value = self._analyzer.OperatingMode.LockIn
        self._analyzer.reference_phase.value = reference_phase
        self._analyzer.input.value = self.input_sources_to_lu_map[input_source]

        self._system_infra.main_in_2_gain.value = self.input_range_to_gain_map[input_range]

        # activate output for modulation
        self.output_setup = self.Output[output.name](self._ll)

        self.measure_time = self._start(
            start_frequency, end_frequency, points, step_time, settle_time,
            float(sweep_amplitude) * (self._analyzer.current_reference_amplitude.value_max)
        )
        return self.measure_time

    def finish_execution(self) -> tuple[np.ndarray, np.ndarray]:
        """ cleanup freq execution and extract data

        result
        -------
            two dimensional array with measured data (complex measurement data, frequency array)
        """
        self._finish()

        # cleanup and restore original setting of output
        del self.output_setup
        self.output_setup = None

        return self._get_active_data()

    def is_executing(self) -> bool:
        """ Polls the measuring status.

        Returns
        -------
        returns True if still measuring. False if finished

        """
        if self.is_busy:
            return (time.time() - self.start_time) <= self.measure_time
        return False

    def get_current_sweep_frequency(self) -> float:
        cur_freq =  self._analyzer.reference_frequency.value
        return cur_freq

    def execute(self, *args, **kwargs):
        """ Do the freq sweep and wait until sweep is finished.
        If blocking is not desired, then use the commands start_execute(), is_executing() and finish_execute()

        Parameters
        ----------
        See start_execute()

        Result
        ------
        see result of finish_execute()
        """

        total_time = self.start_execute(*args, **kwargs)

        print(f"Wait for {total_time:.1f}s.")
        while self.is_executing():
            time.sleep(0.1)

        return self.finish_execution()

    def bode_plot(self, complex_transfer: np.ndarray, frequencies: np.ndarray):
        """Plots Frequency sweep data to Bode Plot.

        complex_transfer
            transfer function as complex numbers
        frequencies
            frequencies where the complex transfer function is given.
        """
        plt.figure()

        plt.subplot(2, 1, 1)
        plt.plot(frequencies, np.abs(complex_transfer))
        plt.title('2nd Lock-In Amplitude - Frequency Sweep')
        plt.xlabel('Frequency [Hz]')
        plt.ylabel('Amplitude V')

        plt.subplot(2, 1, 2)
        plt.plot(frequencies,
            np.unwrap(np.angle(complex_transfer, deg=True), discont=180))
        plt.title('2nd Lock-In Phase - Frequency Sweep')
        plt.xlabel('Frequency [Hz]')
        plt.ylabel('Phase shift [°]')

        plt.tight_layout()
        plt.show()

if __name__ == "__main__":
    """This test requires hardware:
    C3000: Connect cable connectors *Out B* and *User In 1*.
    CX: Connect cable connectors *Fast Out* and *User Input 1*.
    """
    spm_ctrl = nanosurf.SPM()
    spm_ctrl.application.Visible = True
    ll = spm_ctrl.lowlevel
    my_fs = FrequencySweep(spm_ctrl)

    # test if all output configurations work
    for name, output in my_fs.Output.items():
        try:
            my_output = output(ll)
        except:
            print("problem with output : " + name + ":")
            raise
        else:
            del my_output

    data = my_fs.execute(
        start_frequency=100e3, end_frequency=1e6,
        frequency_step=20e3, sweep_amplitude=0.3,
        input_source=InputSource.UserIn1,
        input_range=InputRanges.Full,
        mixer_bw_select=Bandwidths.Hz_360,
        reference_phase=0.0,
        output=FrequencySweepOutput.FastUser_OutB
    )
    my_fs.bode_plot(data[0], data[1])



"""Package for scripting the Nansurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import logging
import nanosurf.lib.spm.com_proxy as spm

class Scanhead():
    def __init__(self, spm: spm.Spm = None, *args):
        """Base class for dealing with Nanosurf scan heads. In this base class general scan head control. 
           For head specific functions, see the corresponding child classes
        """
        self.connected = False
        self.spm = None
        self.spm_app = None
        self.logger = logging.getLogger("Scanhead")

        if spm is not None:
            self.connect(spm)

    def connect(self, spm: spm.Spm) -> bool:
        """ Connect and setup d scan head"""
        if not self.connected:
            if spm is not None:
                self.spm = spm

                if self.spm.is_scripting_enabled():
                    self.spm_app = self.spm.application
                    
                    # let subclass connect to its hardware component
                    self.connected = self.do_connect()
                else:
                    self.logger.error("Scripting interface of spm controller is not enabled.")
            else:
                self.logger.error("Cannot connect to spm controller. spm of None was provided")
        return self.connected

    def is_connected(self) -> bool:
        """ return True if scan head controll is available"""
        return self.connected

    def do_connect(self) -> bool:
        """ This function has to be overwritten by the sub class"""
        raise NotImplementedError(self.__class__.__name__ + '.do_connect')




""" Defines the configuration values of the module.
    Values in the ProStore are saved and loaded during application startup/shutdown automatically
Copyright Nanosurf AG 2021
License - MIT
"""

import enum
import nanosurf.lib.datatypes.sci_val as sci_val
import nanosurf.lib.datatypes.prop_val as prop_val

class DemoSettings(prop_val.PropStore):
    """ settings defined here as PropVal are stored persistently in a ini-file
        settings with a '_' as first char are not stored
    """
    def __init__(self):
        self.repetitions = prop_val.PropVal(int(300))
        self.time_per_repetition = prop_val.PropVal(sci_val.SciVal(0.1, "s"))
        self.send_ticks = prop_val.PropVal(False)
        self.plot_func_id = prop_val.PropVal(int(0))
        self._this_is_not_stored = prop_val.PropVal(int(0))

class DemoResults():
    """ This class saves the worker task result (e.g be read by gui elements or saved to file """
    def __init__(self):
        self.number_of_data_points = 0
        self.last_data = 0.0
        self.mean_value = 0.0

""" Combo boxes show entries with IDs and the names in second list"""
class PlotStyleID(enum.IntEnum):
    PlotSin = 0,
    PlotCos = 1,



""" Multithreading made simpler. 
This package provides classes to handle background tasks and sig/slot communication
Copyright Nanosurf AG 2022
License - MIT
"""
import sys
import time
import logging
from PySide2.QtCore import QThread, QObject, Signal

global_is_in_debugging_mode = (getattr(sys, 'gettrace', lambda : None)() is not None)
if global_is_in_debugging_mode:
    import debugpy 

def activate_debugger_support_for_this_thread():
    """ This function has to be called from each new thread to activate debugger support for it"""
    if global_is_in_debugging_mode:
        debugpy.debug_this_thread()

def slot(f):
    """ decorator which makes debugging possible in slots called in background tasks"""
    def decorated(*args, **kwargs):
        activate_debugger_support_for_this_thread()
        return f(*args, **kwargs)
    return decorated

class NSFThread(QObject):
    """ This class implements basic infrastructure for background activity. 

        The communication with the threads functionality is done by signal/slot mechanism of Qt. 
        Subclass from this class and define signal handler functions. 
        Each handler function should be decorated with @nsf_thead.slot

        Usage
        -----

        Define command signal handler with type Signal() as class attributes
        Redefine _init_in_new_thread() and connect signal handles functions to each defined signal there.

        At runtime, create an instance of your NSFThread class and call start_tread().
        Then emit your defined signal to send commands to your thread

        start_thread() - starts the background task. do_on_start_thread() can implement initialization work
        stop_thread()  - stops the background task. do_on_finish_thread can implement  cleanup work
        is_thread_running() - returns True if the task is running. Even if no worker is executing

        Signals
        -------

        sig_task_started - is sent after background task is started and do_on_start_thread() is executed
        sig_task_finished - is sent after background task  do_on_finish_thread() is executed and background task is stopped

    """
    sig_task_started = Signal()
    sig_task_finished = Signal()

    #--------------- for sub classing ------------------------------------------------------------

    def do_on_start_thread(self):
        pass

    def do_on_finish_thread(self):
        pass

    #--------------- implementation ------------------------------------------------------------

    def __init__(self, *args, **kwargs):
        """ setup the thread and wait until the task is started by thread.start()"""
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)
        self.is_thread_start_done = False
        self.is_thread_finish_done = False
        self.the_thread = QThread()
        self.moveToThread(self.the_thread)
        self.the_thread.started.connect(self._on_thread_started)
        self.the_thread.finished.connect(self._on_thread_finished)
        self._init_in_new_thread()

    def _init_in_new_thread(self):
        """ Init new thread functions here 
        Most common is to connect signal handler functions to signals
        """
        pass
    
    @slot
    def _on_thread_started(self):
        self.logger.info(f"on_thread_started()") 
        self.do_on_start_thread()
        self.is_thread_start_done = True
        self.is_thread_finish_done = False
        self.sig_task_started.emit()
        
    @slot
    def _on_thread_finished(self):
        self.logger.info(f"on_thread_finished()") 
        self.do_on_finish_thread()
        self.is_thread_start_done = False
        self.is_thread_finish_done = True
        self.sig_task_finished.emit()
        
    def start_thread(self):
        """ Start the worker activity. """
        if not self.is_thread_running():
            self.the_thread.start()
        else:
            self.logger.warning("background thread already started")

    def stop_thread(self, wait: bool = False, timeout : float = 10.0) -> bool:
        """ Send stop request to worker thread.  
            'sig_finished' is emitted when the worker thread has finished its job.

            Parameters
            ----------
            wait: bool
                If True the function waits until the worker thread has really finished
            timeout: float
                maximal wait time 

            Returns
            -------
            Bool:
                True if task could be stopped.
        """
        if self.is_thread_running():
            self.logger.info(f"Stop background task")
            self.the_thread.exit(0)
            if wait:
                has_finished = self.the_thread.wait(timeout*1000)
                if not(has_finished):
                    self.logger.warning(f"Thread stop request was not served in {timeout}s")
                    self.the_thread.terminate()
        else:
            self.logger.info("Thread is not running")
        return not self.is_thread_running()

    def is_thread_running(self) -> bool:
        return self.the_thread.isRunning() if self.the_thread is not None else False

class NSFBackgroundWorker(NSFThread):
    """ This class provides a background worker function which can be called multiple times. 
        
        Usage
        -----

        Implement self.do_work() which your code. 
        Call start_thread() once to start background task
        Call start_worker(args, kwargs) as many times as you like 

        New functions
        -------------

        start_work()   - starts the background working task implemented by do_work()
        abort_worker() - send an abort request to the worker implemented in do_work().

        do_work() - implement this function and do the background work. 
                    Check regularly self.is_stop_request_pending()  and abort if it returns True.
                    in 'self._args' and 'self._kwargs' the parameters are passed from start_work()

        is_worker_running() - returns True if the background worker is executing
        is_worker_aborted() - returns True if the worker is finished by abortion

        New Signals
        -----------

        sig_worker_started  - is send each time the background worker function has started to do its work.
        sig_worker_finished - is send each time the background worker function has finished. Also in case of abortion.

    """
    sig_worker_started = Signal()
    sig_worker_finished = Signal()

    _cmd_start_worker = Signal()

    #--------------- to sub class ------------------------------------------------------------
   
    def do_work(self):
        self.logger.info(f"base implementation of do_work(): was called. This function should be overwritten by subclass to do real stuff.") 

    #--------------- implementation ------------------------------------------------------------

    def __init__(self, *args, **kwargs):
        self.worker_stop_request_flag = False
        self.is_worker_active_flag = False
        super().__init__(*args, **kwargs)

    def _init_in_new_thread(self):
        self._cmd_start_worker.connect(self._start_background_worker)
    
    def do_on_start_thread(self):
        self.is_worker_active_flag = False
        self.worker_stop_request_flag = False
        
    def do_on_finish_thread(self):
        self.is_worker_active_flag = False
        
    def start_worker(self, *args, **kwargs):
        """ start the provided function  in the context of the background thread
        sig_function_started is emitted when the function really starts.
        sig_function_finished is emitted when the function returned
        To stop the function, call abort_worker() to try to abort the function.
        The function has to monitor if self.is_stop_request_pending() is True and abort accordingly
        """
        self.worker_stop_request_flag = False
        self.is_worker_active_flag = True
        self._args = args
        self._kwargs = kwargs
        self._cmd_start_worker.emit()

    def abort_worker(self, wait: bool = True, timeout: float = 10.0) -> bool:
        """ Send stop request to client function thread.  

            Parameters
            ----------
            wait: bool
                If True the function waits until the client function has really finished
            timeout: float
                maximal wait time 

            Returns
            -------
            Bool:
                True if function could be stopped.
        """
        if self.is_worker_running():
            self.logger.info(f"Set worker_stop_request_flag")
            self.worker_stop_request_flag = True

            if wait:
                ticks = 0
                waiting_time = 0.0
                while self.is_worker_running() and (waiting_time < timeout):
                    time.sleep(0.1)
                    ticks += 1
                    if ticks > 10:
                        ticks = 0
                        waiting_time += 1.0
                        if waiting_time >= timeout:
                            self.logger.warning(f"Stop worker request was not served in {timeout}s")
                            break
        return not self.is_worker_running()

    def is_worker_running(self) -> bool:
        return self.is_worker_active_flag

    def is_worker_aborted(self) -> bool:
        return self.worker_stop_request_flag and (not self.is_worker_active_flag)

    def stop_thread(self, wait: bool = False, timeout : float = 10.0) -> bool:
        """ Send stop request to worker thread.  
            'sig_finished' is emitted when the worker thread has finished its job.

            Parameters
            ----------
            wait: bool
                If True the function waits until the worker thread has really finished
            timeout: float
                maximal wait time 

            Returns
            -------
            Bool:
                True if task could be stopped.
        """
        if self.is_thread_running():
            self.the_thread.requestInterruption()
            self.worker_stop_request_flag = True

            if self.is_worker_running():
                self.logger.info(f"Stop running worker")
                has_finished = self.abort_worker(wait, timeout)
                if not has_finished:
                    self.logger.info(f"Could not stop worker")

        return super().stop_thread(wait=wait, timeout=timeout)    

    def is_stop_request_pending(self):
        return self.the_thread.isInterruptionRequested() or self.worker_stop_request_flag

    @slot
    def _start_background_worker(self):
        self.is_worker_active_flag = True
        self.sig_worker_started.emit()

        if not self.worker_stop_request_flag:
            self.do_work()
        
        self.is_worker_active_flag = False
        self.sig_worker_finished.emit()



"""Package for scripting the Nanosurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""


import enum
import time
import psutil
import win32com.client

import nanosurf.lib.spm.lowlevel as ll
import nanosurf.lib.spm.lowlevel.ctrlunits.factory as ctrl_factory

_known_spm_app_names = [ 'MobileS', 'SPM_S', 'Easyscan2', 'C3000', 'Naio', 'USPM', 'CoreAFM', 'CX']


class Spm():
    """Base class for dealing with Nanosurf SPMs."""
    
    class SystemState(enum.IntEnum):
        Uncal = 0
        Idle = 1
        Approach = 2
        Scan = 3
        Spec = 4
        Litho = 5
        MacroCmd = 6

    class SystemEnvironment(enum.IntEnum):
        Air = 0
        Liquid = 1

    class SystemHealthState(enum.IntEnum):
        OK = 0
        NotResponding = 1
        SimulationMode = 2

    class SystemZAxisIdleMode(enum.IntEnum):
        ZControllerActive = 0
        RetractTip = 1
        KeepLastPos = 2
        AbsolutePos = 3

    class SystemXYAxisIdleMode(enum.IntEnum):
        ImageCenter = 0
        KeepLastPos = 1

    class SystemMotorID(enum.IntEnum):
        Approach = 0
        FootA = 1
        FootB = 2
        FootC = 3
        Focus = 4
        PTEX = 5
        PTEY = 6
        BeamDefX = 7
        BeamDefY = 8
        PhotoDetLateral = 9
        PhotoDetNormal = 10
        LensGimbal = 11

    class SystemMotorDirection(enum.IntEnum):
        Positive = 0
        Negative = 1

    class SystemMotorSpeedLevel(enum.IntEnum):
        VerySlow = 0
        Slow = 1
        Normal = 2
        Fast = 3
        VeryFast = 4

    class ScanAxisRangeModes(enum.IntEnum):
        Off = 0
        FullRange = 1
        ReducedRange = 2

    class XYClosedLoopModes(enum.IntEnum):
        OpenLoop = 0
        ClosedLoop = 1

    class ZClosedLoopModes(enum.IntEnum):
        FixDrive = 0
        FixPosition = 1

    class VideoSource(enum.IntEnum):
        SideView = 0
        TopView = 1

    class ThermalTuneAvgType(enum.IntEnum):
        ExpDecay = 0
        PropWeight = 1

    class ThermalTuneNsfFit(enum.IntEnum):
        Damping = 0
        Sigma = 1
        ResFreq = 2
        QualitFactor = 3
        ResPeakAmpAboveNoise = 4
        NumOfParams = 5

    class ThermalTuneSHOFit(enum.IntEnum):
        WhiteNoise = 0
        PinkNoise = 1
        ResFreq = 2
        QualitFactor = 3
        ResPeakAmpAboveNoise = 4
        NumOfParams = 5

    class OperatingMode(enum.IntEnum):
        User = 0
        STM = 1
        StaticAFM = 2
        DynamicAFM = 3
        PhaseContrast = 4
        ForceModulation = 5
        SpreadingResistivity = 6
        ContPhase = 7
        DeltaF = 8
        LateralForce = 9

    class ApproachStatus(enum.IntEnum):
        Standby = 0
        Initializing = 1
        Approaching = 2
        ApproachDone = 3
        ApproachAborted = 4
        MoveToParkPos = 5
        ParkPosReached = 6
        MoveAway = 7
        MoveToward = 8
        SensorFailed = 9
        LimitFailed = 10
        CalibrationFailed = 12
        UserAbort = 13
        MaxOut = 14
        Initdone = 15
        AdjustingTipPos = 16

    class ApproachMotorStatus(enum.IntEnum):
        Fail = ord('6')
        Error = ord('5')
        NC = ord('4')
        MaxOou = ord('3')
        MinIn = ord('2')
        InRange = ord('1')
        NotDefined = ord('0')
        Unknown = -1

    class AFMSensorStatus(enum.IntEnum):
        ToLow = ord('7')
        Fail = ord('4')
        ToHigh = ord('3')
        OK = ord('1')
        NotDefined = -1

    class DataFilter(enum.IntEnum):
        RAW = 0
        MeanFit = 1
        LineFit = 2
        DerivedData = 3
        ParabolaFit = 4
        PolynomialFit = 5

    class CantileverProperty(enum.IntEnum):
        LeverLength = 0
        LeverWidth = 1
        SpringConst = 2
        AirResonanceFrq = 3
        AirQFactor = 4
        LiquidResonanceFrq = 5
        LiquidQFactor = 6

    class CalibrationSignalID(enum.IntEnum):
        XAxis = 0
        YAxis = 1
        ZAxis = 2
        TipCurrent = 3
        TipVoltage = 4
        Ch0_Deflection = 5
        Ch0_Amp = 6
        Ch0_Pase = 7
        Ch0_Excitation = 8
        UserADC0 = 9
        UserADC1 = 10
        UserDAC0 = 11
        UserDAC1 = 12
        UserADC2 = 13

    class ZControllerLoopMode(enum.IntEnum):
        Run = 0
        Freeze = 1
        StopAndClear = 2

    class DataConversion(enum.IntEnum):
        Binary16 = 0
        Physical = 1
        Binary32 = 2

    class DataLineFlagMask(enum.IntEnum):
        DataValid = 1
        CurrentData = 2

    class ChartType(enum.IntEnum):
        LineGraph = 0
        ColorMap = 1
        View3D = 2
        ShadedColorMap = 3
        DualLineGraph = 4
        XYLineGraph = 5

    class WindowStyle(enum.IntEnum):
        Hide = 0
        Normal = 1
        Minimized = 2
        Maximized = 3
        ShowNoActive = 4
        ShowActive = 5

    class ScanLineMode(enum.IntEnum):
        Standard = 0
        ConstHeight = 1

    class LineScanningMode(enum.IntEnum):
        Standard = 0
        DualScan = 1
        Interlaced = 2
        SecondScanOnly = 3

    class ScanMeasureMode(enum.IntEnum):
        Undefined = 0
        Forward = 1
        Backward = 2
        ForwardBackward = 3

    class ScanMode(enum.IntEnum):
        Continouse = 0
        ContUp = 2
        ContDown = 3

    class ScanGroupID(enum.IntEnum):
        Forward = 0
        Backward = 1

    class ScanFrameDir(enum.IntEnum):
        NotScanning = 0
        Up = 1
        Down = 2

    class ScanChannel(enum.IntEnum):
        Deflection_or_ZCtrlIn = 0
        Topography = 1
        
    class SpecChannel(enum.IntEnum):
        Deflection_or_ZCtrlIn = 0
        Topography = 1
        
    class TipSignalMode(enum.IntEnum):
        CurrentSensinput = 0
        VoltageOutput = 1
        DirectFeedthrough = 2

    class LaserPowerMode(enum.IntEnum):
        Unidefined = 0
        LaserDrive = 1
        LaserPower = 2
        DetectorSensitivity = 3

    class ExcitationSource(enum.IntEnum):
        Internal = 0
        External = 1

    class ExcitationMode(enum.IntEnum):
        PiezoElectric = 0
        PhotoThermal = 1

    class User0InputPol(enum.IntEnum):
        Positive = 0
        Negative = 1

    class ScanHeadID(enum.IntEnum):
        NC = 0
        Unknown = 1
        NaioSTM = 2
        NaioAFM = 9
        NaniteAFM = 12
        FlexAFM = 14
        LensAFM = 15
        DriveAFM = 18
        CoreAFM = 19
        DriveNMA = 20
        DriveMount = 21

    class FrqSweepResult(enum.IntEnum):
        PeakNotFound = 0
        PeakFound = 1
        Running = 2

    class LithoMode(enum.IntEnum):
        User = 0
        STM = 1
        StaticAFM = 2
        DynamicAFM = 3

    class LithoPenMode(enum.IntEnum):
        LiftTip = 0
        ChangeOpMode = 1

    class LeverExcitationMode(enum.IntEnum):
        InternalSource = 0
        ExternalSource = 1

    class DeflectionUnitMode(enum.IntEnum):
        Volt = 0
        Meter = 1
        Newton = 2

    class SpecRepetitionMode(enum.IntEnum):
        List = 0
        Position = 1

    class SpecEndMode(enum.IntEnum):
        KeepLastZPos = 0
        Approached = 1

    class SpecModulatedOutput(enum.IntEnum):
        ZAxis = 0
        TipVoltage = 1
        UserOut1 = 2
        UserOut2 = 3

    class SpecModulationMode(enum.IntEnum):
        FixedLength = 0
        StopByValue = 1

    class SpecStopMode(enum.IntEnum):
        IsLessThan = 0
        IsGreaterThan = 1

    class SpecPauseMode(enum.IntEnum):
        KeepLastZPos = 0
        ZControllerActive = 1

    class SpecModuleLevel(enum.IntEnum):
        Standart = 0
        Advanced = 1

    class SpecCurrentModulationPhase(enum.IntEnum):
        Idle = 0
        ForwardModulation = 1
        ForwardPause = 2
        BackwardModulation = 3
        BackwardPause = 4

    class SpecGroupNo(enum.IntEnum):
        Forward = 0
        Backward = 1
        ForwardPause = 2
        BackwardPause = 3

    class StageState(enum.IntEnum):
        IdleUnreferenced = 1
        Idle = 2
        Moving = 3
        BackwardPause = 3

    class CantileverGUID():
        AN2_200 = "{BD61D124-8350-4464-BFE4-1D8A156E4913}"
        GLA_1 = "{9E2BA28D-D843-41bf-8F62-05502B3EDB18}"
        ACL_A = "{ABB75273-9543-431a-B681-C79B533DD9E6}"
        ANSCM = "{40AEA787-942C-4d48-A389-DA81571F009C}"
        SICON_A = "{F7A339A7-E29F-42a9-B7AA-D69C54363B76}"
        XYNCHR = "{DD3DFE39-455E-40a1-801E-5D5B14CE4080}"
        XYCONTR = "{12ADC816-C7B1-48f8-8B9E-5E579151CF50}"
        ContAl_G = "{ED5A15E6-D3B0-4e64-8C50-809335D3E143}"
        Multi75E_G = "{9593403B-A476-49a9-AA1F-9C3AEDAC0178}"
        Multi75M_G = "{03D0715C-A520-4976-A5E2-4FC3078E3821}"
        Multi75Al_G = "{443A2EDC-5C9C-4d60-843F-C6688BEA1DEA}"
        Tap190Al_G = "{041FB80E-A179-4170-B5A4-A4EA1CC0A965}"
        Tap150Al_G = "{E0F31C86-6BB8-496b-AC7E-F55C62EAB635}"
        USC_F1_2_k7_3 = "{19AEEE43-478F-4D16-BDB7-2EE256EAF4A4}"
        USC_F0_3_k0_3 = "{16FAEEB6-A887-46F6-A418-81A9EBBCB6C3}"
        Dyn190Al = "{E9CE0D2D-F59E-4B44-A74F-B78C11575E9F}"
        Stat0_2LAuD =  "{A4A16538-CCD1-4BB1-B048-7B4F0F1B31BD}"
        CONTR  = "{89E92173-96FB-4ff9-94D8-42296D00D980}"
        CONTSCR = "{5A687B3E-A75A-4b22-BD70-40ABB931F00E}"
        CONTSCPt  = "{1E95D12B-1DDB-4ace-B3AF-BE9C0D52D4FC}"
        EFMR = "{986305AC-64B5-462e-B37E-6BD5AE447BE3}"
        LFMR = "{C61FCA2C-6D5D-4105-9FDE-640D263E229F}"
        MFMR = "{9499F49F-920F-47ec-80B6-883F683FF056}"
        NCLR = "{62633FD4-0555-4cee-A8B4-B82F4CEFBB48}"
        PPP_FMR = "{EBA2B75C-AA94-4451-AD36-1388CDABF5E8}"
        pq_SCONT = "{8D28AE10-E1DD-49E0-8CC6-ABD7CEDF57B0}"
        qp_CONT = "{0996E3AC-ABF6-4A22-B320-4BF749288156}"
        qp_fast_CB1 = "{3F3DD96B-F838-45B6-AA8C-B54F66ED9571}"
        qp_fast_CB2 = "{964280C3-70F7-4E22-AA60-734E672D7A02}"
        qp_fast_CB3 = "{CCF4B65D-F3D8-4A40-9108-53468ECBA1B4}"

    class ControllerType(enum.IntEnum):
        Undefined = -1
        Sim =       0
        ST9 =       1
        FLEX =      2
        CX =        3

    class FirmwareType(enum.IntEnum):
        Unknown = -1
        FREERTOS = 0
        LINUX = 1

    _class_id = ""  # class_id  The COM class id string, if equal to SPM.Application: try to connect to any running application.
    lowlevel = ll.Lowlevel() # makes code recognition in VisualStudio code possible

    def __init__(self, class_id: str, lu_shared_file_path: str=""):
        """Create the COM client to deal with the SPM.

        Parameters
        ----------
        class_id: str
            COM automation name registered. eg 'CX.Application'
        lu_shared_file_path: str  
            Path to interface definition file, if empty, query the application for the path
        """

        # Connect or start application
        self._class_id = class_id
        self.application = None
        self.lowlevel = None
        self.lu_shared_file_path = lu_shared_file_path

        if self._class_id == "SPM.Application":
            if not self._connect_to_running_app():
                print("Could not find a running Nanosurf SPM application.\nPlease start one first.")
                return
        else:
            try:
                self.application = win32com.client.Dispatch(self._class_id)
            except:
                print("Could not start application", self._class_id)
                
        #  If scripting is enabled finish startup and prepare subclasses
        if self.is_scripting_enabled():
            self._wait_for_end_of_startup()
            self._prepare_lowlevel_scripting()
            
    def is_connected(self) -> bool:
        return self.application is not None

    def is_scripting_enabled(self) -> bool:
        script_enabled = False
        if self.application is not None:
            script_enabled = self.application.Scan is not None
        return script_enabled

    def is_lowlevel_scripting_enabled(self) -> bool:
        self._prepare_lowlevel_scripting()                
        return self.lowlevel is not None
    
    def get_sw_version(self) -> tuple[int,int,int,int]:
        sw_version:str = self.application.Version
        major, minor, revision, bugfix = sw_version.split(".")
        return (int(major), int(minor), int(revision), int(bugfix))

    def get_controller_type(self) -> ControllerType:
        detected_type = Spm.ControllerType.Undefined
        try:
            test_obj = self.application.CreateTestObj
            detected_type = test_obj.GetControllerMode()
        except:
            known_cx_applications = ["CX.Application", "USPM.Application"]
            known_flex_applications = ["C3000.Application", "CoreAFM.Application"]
            if self._class_id in known_cx_applications:
                detected_type = Spm.ControllerType.CX
            elif self._class_id in known_flex_applications:
                detected_type = Spm.ControllerType.FLEX
            else:
                detected_type = Spm.ControllerType.ST9
        return detected_type

    def get_firmware_type(self) -> FirmwareType:
        detect_firmware_type = Spm.FirmwareType.Unknown
        major, minor, *_ = self.get_sw_version()
        if major < 3 or (major == 3 and minor < 4): 
            if self.get_controller_type() != Spm.ControllerType.ST9:
                detect_firmware_type = Spm.FirmwareType.FREERTOS
        else:
            if self.get_controller_type() == Spm.ControllerType.CX:
                detect_firmware_type = Spm.FirmwareType.LINUX
            else:
                detect_firmware_type = Spm.FirmwareType.FREERTOS
        return detect_firmware_type

    def _prepare_lowlevel_scripting(self):
        if self.is_scripting_enabled():
            if self.lowlevel is None and self.application.SPMCtrlManager is not None:
                self.lowlevel = ll.Lowlevel(self.application.SPMCtrlManager, self.lu_shared_file_path)
                self.lowlevel.ctrlunits = ctrl_factory._CtrlUnitFactory(self)

    def _connect_to_running_app(self) -> bool:
        """ try to connect to a running spm application instance """
        self._class_id = ""
        _known_spm_app_names_lowercase = [i.lower() for i in _known_spm_app_names]
        
        for process in psutil.process_iter(['name']):
            procname = str(process.info['name'])
            if procname.endswith('.exe'):
                app_name = procname[:-4]
                if  app_name.lower() in _known_spm_app_names_lowercase:
                    self._class_id = app_name + ".Application"
                    self.application = win32com.client.Dispatch(self._class_id)
                    print(f"Connected to running app: {app_name}")
                    break
        return (self._class_id != "")
        # Should work like this but to whatever reason it gives only a runtime exception:
        # for app_name in _known_spm_app_names:
        #     if self._class_id == "":
        #         try:
        #             self._class_id = app_name + ".Application"
        #             self.application = win32com.client.GetActiveObject(self._class_id)
        #         except:
        #             pass

    def _wait_for_end_of_startup(self) -> bool:
        if self.application.IsStartingUp:
            print("Waiting for controller startup.", end="")
            tick = 0
            while self.application.IsStartingUp and (tick < 30):
                time.sleep(1)
                tick += 1
                print(".", end="")
            print()
            if self.application.IsStartingUp:
                print("Timeout. End of startup not reached.")
        return not self.application.IsStartingUp



# ///
# /// Line Edit for advanced number formatting
# ///
# /// Copyright (C) Nanosurf AG - All Rights Reserved (2021)
# /// Unauthorized copying of this file, via any medium is strictly prohibited
# /// https://www.nanosurf.com
# ///

import typing
import enum
import math
import sys
from PySide2 import QtGui, QtWidgets
from PySide2.QtCore import Qt, QTimer, Signal

import nanosurf.lib.datatypes.sci_val as sci_val
import nanosurf.lib.datatypes.sci_val.unit_prefix as up
import nanosurf.lib.datatypes.sci_val.convert as convert

class StepMode(enum.IntEnum):
    rel_1st_digi = 0, # increment relative the first significant digit 123.45 -> 223.45
    rel_2nd_digi = 1, # increment relative the first significant digit 123.45 -> 223.45
    abs_double = 2,   # increment by doubling the value  123.45 -> 246.9
    abs_01 = 3,       # increment by doubling the value  123.45 -> 246.9
    abs_1 = 4,        # increment by 1
    odd_1 = 5         # increment by 2 to the next odd value

class _SciLineEditQt(QtWidgets.QLineEdit):

    value_changed_event = Signal(float)

    def __init__(self, precision: int = 3, step_mode : StepMode = StepMode.rel_2nd_digi, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._unit_prefix_automode = sci_val.up.Prefix.auto_
        self._precision = precision
        self._precision_default = precision
        self._step_mode = step_mode
        self._step_delay_time = 800
        self._max_characters = 20
        self._block_value_change_signal = False
        self._sci_value = sci_val.SciVal()
        self._edit_finished_timer = QTimer() 
        self._edit_finished_timer.timeout.connect(self._value_changed_timeout)
        self.cursorPositionChanged.connect(self._handle_cursor_position_changed())

    # public interface ------------------------------------------

    def set_value(self, value: typing.Union[sci_val.SciVal, float], notify: bool = True):
        if isinstance(value, sci_val.SciVal):
            self._sci_value._value = value._value
            self._sci_value._unit = value._unit
            self._show_value() 
            if notify:
                self._value_changed()
        else:
            do_notify = notify and (value != self._sci_value.value())
            self._sci_value.set_value(value)
            self._show_value()   
            if do_notify:
                self._value_changed()

    def set_unit(self, unit: str, notify: bool = True):
        do_notify = notify and (unit != self._sci_value.unit())
        self._sci_value.set_unit(unit)
        self._show_value()     
        if do_notify:
            self._value_changed()

    def set_precision(self, precision: int, notify: bool = True):
        do_notify = notify and (precision != self._precision)
        self._precision = precision
        self._show_value()     
        if do_notify:
            self._value_changed()

    def set_prefix_id(self, prefix_id: sci_val.up.Prefix, notify: bool = True):
        do_notify = notify and (prefix_id != self._sci_value.prefix_id())
        self._sci_value.set_prefix_id(prefix_id)
        if do_notify:
            self._value_changed()

    def set_stepping_mode(self, mode: StepMode):
        self._step_mode = mode

    def set_allowed_prefix_ids(self, allowed_prefix_ids: list[up.Prefix]) :
        return self._sci_value.set_allowed_prefixes(allowed_prefix_ids)

    def set_value_min(self, value: float, notify: bool =True):
        oldval = self._sci_value.value()
        self._sci_value.set_value_min(value)
        do_notify = notify and (oldval != self._sci_value.value())
        if do_notify:
            self._value_changed()
              
    def set_value_max(self, value: float, notify: bool = True):
        oldval = self._sci_value.value()
        self._sci_value.set_value_max(value)
        do_notify = notify and (oldval != self._sci_value.value())
        if do_notify:
            self._value_changed()

    def sci_value(self) -> sci_val.SciVal:
        return self._sci_value

    def value(self) -> float:
        return self._sci_value.value()  

    def unit(self) -> str:
        return self._sci_value.unit()  

    def precision(self) -> int:
        return self._precision

    def prefix_id(self) -> sci_val.up.Prefix:
        return self._sci_value.prefix_id()

    def allowed_prefix_ids(self) -> list:
        return self._sci_value.allowed_prefixes()

    def value_min(self) -> float:
        return self._sci_value.value_min()

    def value_max(self) -> float:
        return self._sci_value.value_max()

    def stepping_mode(self) -> StepMode:
        return self._step_mode

    # internal functions --------------------------------------------    

    def focusInEvent(self, event: QtGui.QFocusEvent) -> None:
        """ overwrite handler for widget focus events"""
        super().focusInEvent(event)
        self._limit_cursor_to_last_digi()

    def focusOutEvent(self, event: QtGui.QFocusEvent) -> None:
        """  overwrite handler for widget focus events"""
        super().focusOutEvent(event)
        self._limit_cursor_to_last_digi()
        self._reformat(True)

    def mousePressEvent(self, arg__1: QtGui.QMouseEvent) -> None:
        """ overwrite handler for mouse events"""
        super().mousePressEvent(arg__1)
        if self._is_cursor_behind_last_digit():
            self._limit_cursor_to_last_digi()

    def mouseDoubleClickEvent(self, arg__1: QtGui.QMouseEvent) -> None:
        """ overwrite handler for mouse click events"""
        if self._is_cursor_behind_last_digit():
            start = self._get_last_digi_position()+1
            end = len(self.text())
            self.setCursorPosition(end)
            self.setSelection(start, end -start)
        else:
            super().mouseDoubleClickEvent(arg__1)
            
    def mouseMoveEvent(self, arg__1: QtGui.QMouseEvent) -> None:
        """ overwrite handler for move events"""
        super().mouseMoveEvent(arg__1)
        if self.hasSelectedText():
            last_digi = self._get_last_digi_position()
            start = self.selectionStart()
            end = self.selectionEnd()
            if start <= last_digi:
                if end > (last_digi+1):
                    end = last_digi+1
                    self.setSelection(start, end - start)
            elif start > last_digi:
                end = len(self.text())
                self.setSelection(start, end - start)
  
    def keyPressEvent(self, event: QtGui.QKeyEvent) -> None:
        """ overwrite handler for key events"""
        if self.isReadOnly():
            return
        shift = (event.modifiers() == Qt.ShiftModifier)
        ctrl = (event.modifiers() == Qt.ControlModifier)
        key = event.key()
        if (key == Qt.Key_Return) or (key == Qt.Key_Enter):
            super().keyPressEvent(event)
            self._reformat(True)
        elif (key==Qt.Key_Delete) or event.matches(QtGui.QKeySequence.Delete):
            if not(self._is_cursor_behind_last_digit()):
                super().keyPressEvent(event)
        elif (key == Qt.Key_Escape):
            self._show_value()
            self._reformat(True)
            self.clearFocus()
        elif (key == Qt.Key_Down) or (key == Qt.Key_PageDown):
            self._decrement()
            self._show_value()     
        elif (key == Qt.Key_Up) or (key == Qt.Key_PageUp):
            self._increment()
            self._show_value()   
        elif (key == Qt.Key_F):
            self._update_prefix_id(sci_val.up.Prefix.femto)
        elif (key == Qt.Key_P):
            self._update_prefix_id(sci_val.up.Prefix.pico)
        elif (key == Qt.Key_N):
            self._update_prefix_id(sci_val.up.Prefix.nano)
        elif (key == Qt.Key_M):
            self._update_prefix_id(sci_val.up.Prefix.milli)
        elif (key == Qt.Key_U):
            self._update_prefix_id(sci_val.up.Prefix.micro)
        elif (key == Qt.Key_B) or (key == Qt.Key_Space):
            self._update_prefix_id(sci_val.up.Prefix.base)
        elif (key == Qt.Key_K):
            self._update_prefix_id(sci_val.up.Prefix.kilo)
        elif (key == Qt.Key_M) and shift:
            self._update_prefix_id(sci_val.up.Prefix.mega)
        elif (key == Qt.Key_G):
            self._update_prefix_id(sci_val.up.Prefix.giga)
        elif (key == Qt.Key_T):
            self._update_prefix_id(sci_val.up.Prefix.tera)
        elif (key == Qt.Key_H) and ctrl:
            self._toggle_auto_prefix_mode()
        elif (key == Qt.Key_A) and ctrl:
            self._toggle_auto_prefix_mode()
        elif (key == Qt.Key_Backspace) :
            super().keyPressEvent(event)
        elif event.matches(QtGui.QKeySequence.SelectAll) or event.matches(QtGui.QKeySequence.Copy) or event.matches(QtGui.QKeySequence.Paste):
            super().keyPressEvent(event)
            self._handle_selection()
        elif (key == Qt.Key_Insert) or (key == Qt.Key_Insert) or (key == Qt.Key_End):
            super().keyPressEvent(event)
            self._handle_selection()
        elif (key == Qt.Key_Home):
            self.setCursorPosition(0)
            super().keyPressEvent(event)
        elif (key == Qt.Key_End):
            self.setCursorPosition(len(self.text()))
            self._limit_cursor_to_last_digi()
            super().keyPressEvent(event)
        elif (key == Qt.Key_Left):
            if self.text() == self.selectedText():
                self.setCursorPosition(0)
            super().keyPressEvent(event)
            self._limit_cursor_to_last_digi()
        elif (key == Qt.Key_Right):
            if self.text() == self.selectedText():
                self.setCursorPosition(len(self.text()))
                self._limit_cursor_to_last_digi()
            else:    
              super().keyPressEvent(event)
              self._limit_cursor_to_last_digi()
        elif (key == Qt.Key_0) or \
             (key == Qt.Key_1) or \
             (key == Qt.Key_2) or \
             (key == Qt.Key_3) or \
             (key == Qt.Key_4) or \
             (key == Qt.Key_5) or \
             (key == Qt.Key_6) or \
             (key == Qt.Key_7) or \
             (key == Qt.Key_8) or \
             (key == Qt.Key_9):
            if len(self.text()) < self._max_characters:
                super().keyPressEvent(event)
        elif (key == Qt.Key_Plus) or \
             (key == Qt.Key_Minus):
            if len(self.text()) < self._max_characters:
                super().keyPressEvent(event)
        elif (key == Qt.Key_Asterisk) or \
             (key == Qt.Key_Slash) or \
             (key == Qt.Key_AsciiCircum) or \
             (key == Qt.Key_E) or \
             (key == Qt.Key_Period):
            if len(self.text()) > self._max_characters:
                return
            if self.text() == self.selectedText():
                self.setCursorPosition(len(self.text()))
                self._limit_cursor_to_last_digi()
            super().keyPressEvent(event)
            self._handle_selection()
        else:
            if not (ctrl or shift):
                self._handle_selection()

    def wheelEvent(self, event: QtGui.QWheelEvent) -> None:
        """ overwrite handler for wheel event"""
        if self.isReadOnly():
            return
        has_focus = self.hasFocus()
        delta = event.angleDelta().y()
        if has_focus and (delta >0.0):
            self._increment()
            event.accept()
        elif has_focus and (delta < 0.0):
            self._decrement()
            event.accept()
        else:
            super().wheelEvent(event)    

    def _handle_cursor_position_changed(self):
        self._limit_cursor_to_last_digi()

    def _show_value(self):
        self.setText(self._sci_value.to_string_formatted(prefix_id=self._sci_value.prefix_id(), precision=self._precision))  

    def _calc_step_size(self, is_incrementing:bool):
        if self._step_mode == StepMode.rel_1st_digi:
            return self._calc_step_size_real_significant(0, is_incrementing)    
        elif self._step_mode == StepMode.rel_2nd_digi:
            return self._calc_step_size_real_significant(1, is_incrementing)    
        elif self._step_mode == StepMode.abs_double:
            return self._calc_step_size_double(is_incrementing)    
        elif self._step_mode == StepMode.abs_01:
            return self._calc_step_size_abs(0.1, is_incrementing)                
        elif self._step_mode == StepMode.abs_1:
            return self._calc_step_size_abs(1.0, is_incrementing)                
        elif self._step_mode == StepMode.odd_1:
            return self._calc_step_size_odd(is_incrementing)                            
        return self._calc_step_size_real_significant(1, is_incrementing)

    def _calc_step_size_real_significant(self, sig:int, is_inc:bool) -> float:
        current_val = self._sci_value.value()
        value_abs = abs(current_val) + sys.float_info.epsilon
        number_digits = math.floor(math.log10(value_abs))
        start_with_10 = (math.floor(value_abs / math.pow(10, number_digits-1)) == 10.0)
        dec_pos_num =  (not(is_inc) and current_val >= 0.0)        
        inc_neg_num =  (    is_inc  and current_val < 0.0)

        if  start_with_10 and (dec_pos_num or inc_neg_num):
            sig += 1

        second_significant = number_digits - sig
        step_size = math.pow(10, second_significant)

        if not self._sci_value._is_real:    
            step_size = math.ceil(step_size)
            step_size = min(max(1.0, math.ceil(step_size)), step_size)
        elif step_size == 0.0:
            step_size = up.prefix_id_to_absolute_value(self._sci_value._prefix_id) / math.pow(10, sig)    
        return step_size

    def _calc_step_size_abs(self, step:float, is_inc:bool) -> float:
        current_val = self._sci_value.value()
        if is_inc:
            new_value = current_val + step
        else:
            new_value = current_val - step

        if new_value == 0.0:
            return step + 0.01
        return step

    def _calc_step_size_odd(self, is_inc:bool) -> float:
        current_val = self._sci_value.value()
        step = 2
        if is_inc:
            new_value = current_val + step
        else:
            new_value = current_val - step
        if (new_value % 2) == 0:
            return step + 1
        return step    

    def _calc_step_size_double(self, is_inc:bool) -> float:
        current_val = self._sci_value.value()
        if is_inc:
            return current_val
        return current_val / 2.0    

    def _adapt_step_size(self, step_size:float) ->float:
        if (self._step_mode == StepMode.rel_1st_digi or self._step_mode == StepMode.rel_2nd_digi):
            expo = -1 * int(self._precision)
            min_step = math.pow(10,expo) * up.prefix_id_to_absolute_value(self._sci_value._prefix_id)
            if step_size <= min_step:
                return min_step
        return step_size    

    def _make_odd_if_step_odd1(self, val:float) -> float:
        must_be_odd = (self._step_mode == StepMode.odd_1) and not(self._sci_value._is_real)  
        if must_be_odd:
            is_even = ((val % 2) == 0)                 
            if is_even:
                return val -1

        return val
    def _update_prefix_id(self, new_prefix_id: sci_val.up.Prefix):
        cur_text = self.text() 

        # set cursor at the end of the number
        if self._is_cursor_behind_last_digit():
            display_prefix = new_prefix_id
        else:
            display_prefix = sci_val.up.Prefix.auto_

        self._limit_cursor_to_last_digi()
        cur_pos = self.cursorPosition()     
        
        # add to number ne prefix
        new_text = cur_text[:cur_pos]
        new_text = new_text + sci_val.up.prefix_id_to_string(new_prefix_id)+self.unit()
        res = sci_val.convert.to_value(new_text,self.unit())
        if res.success:
            self._sci_value.set_value(res.value)

        self._sci_value.set_prefix_id(display_prefix)

        self._value_changing()
        self._presentaton_changed()
        self._reposition_icons()   

    def _is_auto_mode(self) -> bool:
        return (self._sci_value.prefix_id() == sci_val.up.Prefix.auto_)

    def _toggle_auto_prefix_mode(self):
        if self._is_auto_mode():
            self._sci_value.set_prefix_id(self._unit_prefix_automode)
            self._highlight_unit_and_prefix()
        else:
            self._sci_value.set_prefix_id(sci_val.up.Prefix.auto_)
            self._set_default_text_format()
            self._presentaton_changed()

    def _increment(self):
        self._reformat(False)
        self._prepare_step_mode()
        self._value_changing(self._step_delay_time)
        step_size = self._calc_step_size(is_incrementing=True)
        step_size = self._adapt_step_size(step_size)
        self.set_value(self._sci_value.value() + step_size)
        self._presentaton_changed()

    def _decrement(self):
        self._reformat(False)
        self._prepare_step_mode()
        self._value_changing(self._step_delay_time)
        step_size = self._calc_step_size(is_incrementing=False)
        step_size = self._adapt_step_size(step_size)
        self.set_value(self._sci_value.value() - step_size)
        self._presentaton_changed()

    def _reformat(self, notify: bool):
        cur_text = self.text()
        has_val_changed = self._sci_value.from_string(cur_text)
        if has_val_changed and notify:
            self._edit_finished_timer.stop()
            self._presentaton_changed()
            self._value_changed()
        else:
            self._presentaton_changed()

    def _value_changing(self, timeout=0):
        self._edit_finished_timer.stop()
        self._edit_finished_timer.start(timeout)

    def _value_changed_timeout(self):
        self._edit_finished_timer.stop()
        cur_text = self.text()
        if not(cur_text=="") and (cur_text[0]!=" "):
            self._reformat(False)
            self._value_changed()

    def _value_changed(self):
        if not(self._block_value_change_signal):
            self.value_changed_event.emit(self._sci_value.value())    

    def _presentaton_changed(self):
        self._show_value()
        self._limit_cursor_to_last_digi()
        if self._is_auto_mode():
            t = sci_val.convert.to_value(self.text(), self._sci_value.unit, self._sci_value.prefix_id())
            if t.success:
                self._unit_prefix_automode = t.prefix_id
        else:
            self._highlight_unit_and_prefix()

    def _reposition_icons(self):
        pass

    def _set_default_text_format(self):
        pass

    def _prepare_step_mode(self):
        pass
    
    def _handle_selection(self):
        txt = self.text()
        prefix_id = self._sci_value.prefix_id()
        if self._is_auto_mode():
            prefix_str = sci_val.up.prefix_id_to_string(self._unit_prefix_automode)
        else:
            prefix_str = sci_val.up.prefix_id_to_string(prefix_id)
        expected_end = prefix_str + self._sci_value.unit()

        if (len(txt) < len(expected_end)):
            txt = txt + expected_end 
            self.setText(txt)
            self._limit_cursor_to_last_digi()
        elif not(txt[-len(expected_end):]==expected_end) and not(txt[-len(self._sci_value.unit())]==self._sci_value.unit()):
            new_txt = txt + expected_end
            self.setText(new_txt)
            self._limit_cursor_to_last_digi()

        if not(self._is_auto_mode()):
            self._highlight_unit_and_prefix()

    def _is_cursor_behind_last_digit(self):
        cur_pos = self.cursorPosition()
        last_digi_pos = self._get_last_digi_position()
        return (cur_pos > last_digi_pos)

    def _limit_cursor_to_last_digi(self):
        cur_pos = self.cursorPosition()
        last_digi_pos = self._get_last_digi_position()

        if cur_pos > last_digi_pos:
            self.setCursorPosition(last_digi_pos+1)     

    def _get_last_digi_position(self) ->int:
        cur_str = self.text()
        last_digi_pos = -1
        for index, c in enumerate(cur_str):
            if (c >= '0' and c <= '9') or (c=='.') or (c==',') or (c=='-'):
                last_digi_pos = index
            else:
                break
        return last_digi_pos

    def _highlight_unit_and_prefix(self):
        prefix_str =  sci_val.up.prefix_id_to_string(self._sci_value.prefix_id()) + self._sci_value.unit()
        prefix_str_len = len(prefix_str)
        self._set_italic_range(len(self.text())-prefix_str_len,prefix_str_len)

    def _set_italic_range(self, start:int, length:int):
        custom_format = QtGui.QTextCharFormat()
        custom_format.setProperty(QtGui.QTextCharFormat.FontItalic,True)
        range = QtGui.QTextLayout.FormatRange()
        range.start = start
        range.length = length
        range.format = custom_format
        self._format_range(range)
        self.update()

    def _format_range(self, fr:QtGui.QTextLayout.FormatRange):
        start = fr.start - self.cursorPosition()
        length = fr.length
        # TODO: send text format to edit: 
        # see lineedit_qt.cpp:226 lineedit_qt::format_range(...)


class NSFSciEdit(QtWidgets.QWidget):
    """ Custom Qt Widget to show scientific numbers in a edit with a descriptive label. """
    value_changed_event = Signal(float)

    def __init__(self, label_str = "", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._setup_widgets(label_str)
        self._edit.value_changed_event.connect(self._on_value_changed_from_edit)
        self.set_prefix_id(sci_val.up.Prefix.auto_)
        self.set_precision(3)
 
    def set_value(self, value: typing.Union[sci_val.SciVal, float], notify: bool = True):
        self._edit.set_value(value, notify)
       
    def set_unit(self, unit_str: str, notify: bool = True):
       self._edit.set_unit(unit_str,notify)
   
    def set_label(self, label: str):
       self._label.setText(label) 

    def set_precision(self, precision: int, notify: bool = True):
        self._edit.set_precision(precision,notify)

    def set_prefix_id(self, prefix_id: sci_val.up.Prefix, notify: bool = True):
        self._edit.set_prefix_id(prefix_id,notify)

    def set_allowed_prefix_ids(self, allowed_prefix_ids: list[sci_val.up.Prefix]) :
        return self._edit.set_allowed_prefix_ids(allowed_prefix_ids)

    def set_read_only(self, set_read_only: bool):
        self._edit.setReadOnly(set_read_only)
        self._edit.setDisabled(set_read_only)

    def set_value_min(self, value: float, notify: bool = True):
       self._edit.set_value_min(value, notify)
               
    def set_value_max(self, value: float, notify: bool = True):
       self._edit.set_value_max(value, notify)
               
    def set_value_min_max(self, value_min: float, value_max: float, notify: bool = True):
       self._edit.set_value_min(value_min, notify)
       self._edit.set_value_max(value_max, notify)
    
    def read_only(self) -> bool:
        return self._edit.isReadOnly()

    def label(self) -> str:
        return self._label.text()

    def value(self) -> float:
        return self._edit.value()  

    def sci_value(self) -> sci_val.SciVal:
        return self._edit.sci_value()  

    def unit(self) -> str:
        return self._edit.unit()  

    def precision(self) -> int:
        return self._edit.precision()

    def prefix_id(self) -> sci_val.up.Prefix:
        return self._edit.prefix_id()

    def allowed_prefix_ids(self) -> list[sci_val.up.Prefix]:
        return self._edit.allowed_prefix_ids()

    def value_min(self) -> float:
        return self._edit.value_min()

    def value_max(self) -> float:
        return self._edit.value_max()

    # internal ----------------------------------------------------
    
    def _setup_widgets(self, label_str: str):
        layout = QtWidgets.QVBoxLayout()
        self._label = QtWidgets.QLabel()
        self._label.setText(label_str)
        self._label.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        layout.addWidget(self._label, alignment=Qt.AlignBottom)
        self._edit = _SciLineEditQt()
        self._edit.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        layout.addWidget(self._edit, alignment=Qt.AlignTop)
        self.setLayout(layout)

    def _on_value_changed_from_edit(self, newvalue: float):
        self.value_changed_event.emit(newvalue)




""" The application entry point
Copyright Nanosurf AG 2021
License - MIT
"""

import sys
import nanosurf as nsf

from app import app_base
from modules.scan_module import module as scan_func
from modules.scan_module import gui as scan_gui

MyCompany = "Nanosurf"
MyAppNameShort = "PythonScanDemo"
MyAppNameLong = "Nanosurf Python Imaging and Library Demo"

class MyAppSettings(app_base.AppSettings):
    """ Settings defined here as PropVal are stored persistently in a ini-file"""
    AppHelloMsg = nsf.PropVal("Welcome to scan and library usage demo.")
    
class MyApp(app_base.ApplicationBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.settings = MyAppSettings()

    def do_startup(self):
        """ Insert any module used in this app. 
            If more than one module is added, a menu bar is shown in the app window
        """
        self.add_module(scan_func.ScanModule(self, scan_gui.ScanScreen()), "Scan Demo")
        # use self.add_module() again for second and more modules used in this application

        # here we apply a setting value just for fun
        self.show_message(self.settings.AppHelloMsg.value)
        
    def do_shutdown(self):
        """ Handle cleaning up stuff here"""
        pass

if __name__ == "__main__":
    App = MyApp(MyCompany, MyAppNameShort, MyAppNameLong, [])
    App.start_app()
    sys.exit(App.exec_())




# studio node base class.py

from ast import If
import os
import pathlib
import enum
import re
import lupa
from typing import Any
import hashlib
import importlib
import nanosurf.lib.datatypes.sci_val as sci_val

g_this_compiler_version = "1.0"
g_context_package_name_prefix = "cmd_tree_"
g_cmd_tree_root_name = "root"

class LuaType(enum.Enum):
    Nil = "nil"
    Table = "list"
    Float = "float"
    Int = "int"
    Str = "str"
    Bool = "bool"
    Function = "function"

def get_lua_type(obj) -> LuaType:
    """ Get the type of a lua object as LuaType enum """
    lua_type = lupa.lua_type(obj) if lupa.lua_type(obj) else type(obj)

    if lua_type == "function":
        return LuaType.Function
    elif lua_type == "table":
        return LuaType.Table
    elif lua_type == list:
        return LuaType.Table
    elif lua_type == str:
        return LuaType.Str
    elif lua_type == float:
        return LuaType.Float
    elif lua_type == int:
        return LuaType.Int
    elif lua_type == bool:
        return LuaType.Bool
    return LuaType.Nil

def get_lua_type_str(obj) -> str:
    """ Get the type of a lua object as string representation """
    # if argument is not already a LuaType, try to convert it
    if not(isinstance(obj, LuaType)):
        lua_type = get_lua_type(obj)
    else:
        lua_type = obj
    try:
        return lua_type.value
    except:
        pass
    return LuaType.Nil.value

class CmdTreeNode():
    def __init__(self):
        self._lua_tree_name = ""
        self._context : 'StudioScriptContext' = None
        
    def __getitem__(self, key):
        return self._context.get(f"{self._lua_tree_name}[{key}]")

    def __setitem__(self, key, value):
        return self._context.set(f"{self._lua_tree_name}[{key}]",value)

class CmdTreeProp(CmdTreeNode):
    def __init__(self):
        super().__init__()
        self._lua_value_type = LuaType.Nil

    def __repr__(self) -> str:
        if self._lua_value_type == LuaType.Float or self._lua_value_type == LuaType.Int:
            return sci_val.SciVal(self).to_string()
        return super().__repr__()

class CmdTreeCompiler():
    def __init__(self, parent : 'StudioScriptContext'):
        self.parent_context = parent
        self._file_content = [""]
        self._last_error = ""
        self.my_own_path = pathlib.Path(os.path.dirname(os.path.abspath(__file__)))
        self.regex_non_enum_chars = re.compile('[^0-9a-zA-Z]+', flags=re.UNICODE); 
        self.regex_first_char_a_number = re.compile('\d\w*', flags=re.UNICODE); 

    def build_wrapper_class(self, context_name: str, cmd_table) -> bool:
        self.cmd_package_name = g_context_package_name_prefix + context_name 

        # check if recompilation of command tree is necessary
        runtime_cmd_tree_hash = self.get_table_version(cmd_table, g_cmd_tree_root_name)
        compiled_cmd_tree_hash, compiled_compiler_version = self.get_compiled_versions()

        if (runtime_cmd_tree_hash != compiled_cmd_tree_hash) or (compiled_compiler_version != g_this_compiler_version):
            compiled_tree_file = self.my_own_path / (self.cmd_package_name + '.py')
            if self.compile_command_tree_to_file(cmd_table, runtime_cmd_tree_hash, compiled_tree_file, root_node_name=g_cmd_tree_root_name):
                # after compiling load the new generated wrapper
                mod = importlib.import_module("nanosurf.lib.spm.studio.wrapper."+self.cmd_package_name)
                importlib.reload(mod)
            else:
                self._last_error = f"Could not import dynamically created package from:\n{compiled_tree_file}"
                return False
        return True

    @property
    def last_error(self) -> str:
        return self._last_error

    def get_compiled_versions(self) -> tuple[str, str]:
        tree_hash = ""
        compiler_version = ""
        try:
            mod = importlib.import_module("nanosurf.lib.spm.studio.wrapper."+self.cmd_package_name)
            tree_hash = mod.g_cmd_tree_hash
            compiler_version = mod.g_compiler_version
        except:
            pass
        return (tree_hash, compiler_version)

    def get_table_version(self, table, lua_tree_name: str) -> str:
        hash_str = ""
        hash_str = self._get_table_hash_str(table, lua_tree_name)
        hash_str_sorted = ";".join(sorted(hash_str.split(";")))
        table_hash = hashlib.md5(bytes(hash_str_sorted, encoding='ascii'))
        return table_hash.hexdigest()

    def _get_table_hash_str(self, table, lua_tree_name: str) -> str:
        hash_str = lua_tree_name
        for item_name, item_value in table.items():
            item_type = get_lua_type(item_value) 
            if item_type == LuaType.Table:
                new_lua_tree_name = str(lua_tree_name+"."+item_name)
                hash_str += str(";" + self._get_table_hash_str(item_value, new_lua_tree_name))
            elif (item_type != LuaType.Nil) and type(item_name) == str:
                new_lua_tree_name = str(lua_tree_name+"."+item_name)
                hash_str += str(";" + new_lua_tree_name)
        return hash_str 

    def _has_table_functions(self, table) -> bool:
        for _, item_value in table.items():
            if get_lua_type(item_value) == LuaType.Function:
                return True
        return False

    def _is_property_access_func(self, func_name:str) -> bool:
        suffix = func_name[-4:]
        return (suffix == "_set") or (suffix == "_get")

    def _is_table_a_studio_property(self, table) -> bool:
        """ Detect if a lua table contains a studio property definition
        
        Implementation
        --------------
        This detection method is coupled with the studio lua script "make_property.lua"
        This script implements the metatable implementation to access variables.
        for each variable it create a pair of set/get function.
        """
        found_access_func = False
        found_value = False

        # search for key word
        for item_name, _ in table.items():
            if item_name == "value":
                found_value = True  
                break

        # if key word found check for accessor function
        if found_value:
            for item_name, item_value in table.items():
                item_type = get_lua_type(item_value)
                if (item_type == LuaType.Function) and self._is_property_access_func(item_name):
                    found_access_func = True
                    break

        # if key value and accessor found, it is a property definition
        return found_access_func and found_value 

    def _is_enum_property(self, prop_elements:dict[str, str]) -> bool:
        return "enum" in prop_elements
        
    def _convert_enum_str_to_enum_class_name(self, enum_name:str) -> str:
        class_name = enum_name
        class_name = self.regex_non_enum_chars.sub('_', class_name)
        if self.regex_first_char_a_number.match(class_name):
            class_name =  "num_" + class_name
        return class_name

    def _get_correct_number_type(self, lua_tree_name : str,  item_name : str) -> LuaType:
        type_str = self.parent_context.lua_number_type_str(f"{lua_tree_name}.{item_name}")
        if type_str == "float":
            return LuaType.Float
        elif type_str == "integer":
            return LuaType.Int
        return LuaType.Nil        

    def _get_studio_property_elements(self, prop_table, lua_tree_name : str) -> dict[str, str]:
        """ Extract from property style lua_table a list of all property names and their types
            returns a dict with key:name, val:type
        """
        property_elements: dict[str, str] = {}
        for item_name, item_value in prop_table.items():
            item_type = get_lua_type(item_value)
            if item_type == LuaType.Int:
                item_type = self._get_correct_number_type(lua_tree_name, item_name)
            if item_type != LuaType.Function:
                property_elements[item_name] = get_lua_type_str(item_type)
        return property_elements

    def compile_command_tree_to_file(self, studio_cmd_tree, table_version_hash: str, python_file: pathlib.Path, root_node_name:str) -> bool:
        done = False

        # create file content
        self._file_content = [""]
        self._write_cmd_tree_node(studio_cmd_tree, class_name=root_node_name.capitalize(), lua_tree_name=root_node_name)
        self.write_header(table_version_hash, g_this_compiler_version)

        # write content to file
        try:
            f =  open(python_file, "w")    
        except:
            self._last_error = "could not create file: '{python_file}'"
            return done
        
        try:
            for l in self._file_content:
                f.write(l)
            done = True
        except:
            self._last_error = "could not write to file: '{python_file}'"

        f.close()
        return done

    def create_pretty_print_command_table(self, table) -> str:
        self.dump = ""
        self._dump_command_table(table, indent="")
        return self.dump

    def _dump_command_table(self, table, indent:str) -> str:
        for item_name, item_value in table.items():
            item_type = get_lua_type(item_value)
            self.dump += f"{indent}{item_name}={get_lua_type_str(item_type)}"
            if item_type == LuaType.Table:
                self._dump_command_table(item_value, indent + "   "  )

    def write_header(self, table_version_hash: str, compiler_version: str):
        class_list = [] 
        class_list.append(f"# studio_wrapper.py\n")
        class_list.append(f"\n")
        class_list.append(f"from enum import Enum\n")
        class_list.append(f"from typing import Any\n")
        class_list.append(f"import nanosurf.lib.spm.studio.wrapper as wrap\n")
        class_list.append(f"\n")
        class_list.append(f"g_cmd_tree_hash = '{table_version_hash}'\n")
        class_list.append(f"g_compiler_version = '{compiler_version}'\n")
        class_list.append(f"\n")
        
        self._file_content = class_list + self._file_content

    def _write_cmd_tree_node(self, table, class_name:str, lua_tree_name: str) -> list[str]:
        class_list = []

        if self._is_table_a_studio_property(table):
            class_list.append(f"class {class_name}(wrap.CmdTreeProp):\n")
            prop_elements = self._get_studio_property_elements(table, lua_tree_name)

            is_prop_enum_type = self._is_enum_property(prop_elements)
            if is_prop_enum_type:
                enum_table = table["enum"]
                class_list.append("\n")
                class_list.append(f"    class EnumType(Enum):\n")
                for _, enum_name in enum_table.items():
                    class_list.append(f"        {self._convert_enum_str_to_enum_class_name(enum_name)} = '{enum_name}'\n")
                class_list.append("\n")

            class_list.append(f"    def __init__(self, context: 'StudioScriptContext'):\n")
            class_list.append(f"        super().__init__()\n")
            class_list.append(f"        self._context = context\n")
            class_list.append(f"        self._lua_tree_name = '{lua_tree_name}'\n")
            class_list.append(f"        self._lua_value_type = wrap.LuaType('{prop_elements['value']}')\n")
            class_list.append("\n")

            for value_name, value_type in prop_elements.items():
                if (value_name == "value") and is_prop_enum_type:
                    value_type = "EnumType"
                    class_list.append(f"    @property\n")
                    class_list.append(f"    def {value_name}(self) -> {value_type}:\n")
                    class_list.append(f"        return {class_name}.{value_type}(self._context.get('{lua_tree_name}.{value_name}'))\n")
                    class_list.append("\n")
                    class_list.append(f"    @{value_name}.setter\n")
                    class_list.append(f"    def {value_name}(self, new_val:{value_type}):\n")
                    class_list.append(f"        self._context.set('{lua_tree_name}.{value_name}', new_val.value)\n")
                    class_list.append("\n")
                else:
                    class_list.append(f"    @property\n")
                    class_list.append(f"    def {value_name}(self) -> {value_type}:\n")
                    class_list.append(f"        return {value_type}(self._context.get('{lua_tree_name}.{value_name}'))\n")
                    class_list.append("\n")
                    class_list.append(f"    @{value_name}.setter\n")
                    class_list.append(f"    def {value_name}(self, new_val:{value_type}):\n")
                    class_list.append(f"        self._context.set('{lua_tree_name}.{value_name}', {value_type}(new_val))\n")
                    class_list.append("\n")

            class_list.append("\n")

            self._file_content = class_list + self._file_content
        else:
            class_list.append(f"class {class_name}(wrap.CmdTreeNode):\n")
            class_list.append(f"    def __init__(self, context: 'StudioScriptContext'):\n")
            class_list.append(f"        super().__init__()\n")
            class_list.append(f"        self._context = context\n")
            class_list.append(f"        self._lua_tree_name = '{lua_tree_name}'\n")

            for item_name, item_value in table.items():
                item_type = get_lua_type(item_value)
                if item_type == LuaType.Table:
                    class_list.append(f"        self.{item_name} = {class_name+item_name.capitalize()}(self._context)\n")
            class_list.append(f"\n")

            for item_name, item_value in table.items():
                item_type = get_lua_type(item_value)
                if item_type == LuaType.Function:
                    if not(self._is_property_access_func(item_name)):
                        class_list.append(f"    def {item_name}(self, *args) -> Any:\n")
                        class_list.append(f"        return self._context.call('{lua_tree_name}.{item_name}', *args)\n")
                        class_list.append(f"\n")
                elif (item_type != LuaType.Nil) and (item_type != LuaType.Table) and type(item_name) == str:
                    if item_type == LuaType.Int:
                        item_type = self._get_correct_number_type(lua_tree_name, item_name)
                    type_str = get_lua_type_str(item_type)
                    class_list.append(f"    @property\n")
                    class_list.append(f"    def {item_name}(self) -> {type_str}:\n")
                    class_list.append(f"        return {type_str}(self._context.get('{lua_tree_name}.{item_name}'))\n")
                    class_list.append("\n")
                    class_list.append(f"    @{item_name}.setter\n")
                    class_list.append(f"    def {item_name}(self, new_val:{type_str}):\n")
                    class_list.append(f"        self._context.set('{lua_tree_name}.{item_name}', {type_str}(new_val))\n")
                    class_list.append("\n")        

            class_list.append("\n")

            # add new class on top of base classes
            self._file_content = class_list + self._file_content

            # follow cmd_tree nodes recursively
            for item_name, item_value in table.items():
                item_type = get_lua_type(item_value)
                if item_type == LuaType.Table:
                    new_class_name = str(class_name+item_name.capitalize())
                    new_lua_tree_name = str(lua_tree_name+"."+item_name)
                    self._write_cmd_tree_node(item_value, new_class_name, new_lua_tree_name)



"""Package for scripting the Nanosurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import time
import logging
import enum
import typing
import nanosurf.lib.spm.com_proxy as spm

class MotorID():
    """ Defines the identification parameters of a direct motor control motor. 
        Has to be constructed with initial parameters for each motor. Normally the scan head class is doing so.
    """
    def __init__(self, proxy_index: int, lu_index: int, range: float):
        self.ProxyIndex = proxy_index
        self.LUIndex = lu_index
        self.Range = range        

class MoveStatus(enum.IntEnum):
    """ Defines motor move status"""
    Undefined = 0,
    IdleUnreferenced = 1,
    IdleReferenced = 2,
    Referencing = 3,
    Moving = 4

class LimitStatus(enum.IntEnum):
    """ Defines motor limit switch status"""
    Undefined = 0
    InNoLimit = 1,
    InUpperLimit = 2,
    InLowerLimit = 3,
    InBothLimits = 4, 
    InMultipleLimits = 5

class Direction(enum.IntEnum):
    """ Defines motor move directions for various functions """
    Forward = 0,
    Backward = 1

class MotorState():
    def __init__(self):
        # Axis Status
        self.move_state = MoveStatus.Undefined
        self.limit_state = LimitStatus.Undefined

class MotorControl:
    """ Base class for dealing with Nanosurf Motors. In this base class general motor control functions are used. 
        For measurement head specific functions, see the corresponding child classes
    """

    class _MoveMode(enum.IntEnum):
        Relative = 0,
        Absolute = 1

    def __init__(self, spm: spm.Spm = None, *args):
        """Base class for dealing with Nanosurf Motors. In this base class general motor control functions are used. 
           For measurement head specific functions, see the corresponding child classes
        """
        self.connected = False
        self.spm = None
        self.spm_app = None
        self.spm_system = None
        self.lu_direct_motor_ctrl = None
        self.logger = logging.getLogger("Motor_Control")
        self.tick_callback = self.default_tick_callback

        self.__sleep_time_after_motor_command = 0.2 # seconds
        self.__motor_speed = 1.0 # full speed
        self.__status_poll_time = 0.0
        self.__max_wait_time = 20.0 # [s]

        self.map_move_status = [
            MoveStatus.IdleUnreferenced,
            MoveStatus.IdleReferenced,
            MoveStatus.Referencing,
            MoveStatus.Moving
        ]
        self.map_limit_status = [
            LimitStatus.InNoLimit,
            LimitStatus.InUpperLimit,
            LimitStatus.InLowerLimit,
            LimitStatus.InBothLimits, 
            LimitStatus.InMultipleLimits
        ]

        if spm is not None:
            self.connect(spm)

    def connect(self, spm: spm.Spm) -> bool:
        """ Connect and setup direct motor control of embedded motors"""
        if not self.connected:
            if spm is not None:
                self.spm = spm
                if self.spm.is_scripting_enabled():
                    self.spm_app = self.spm.application
                    self.spm_system = self.spm_app.System
                    if self.spm.is_lowlevel_scripting_enabled():
                        self.lu_direct_motor_ctrl = self.spm.lowlevel.DirectMotorControl(self.spm.lowlevel.DirectMotorControl.Instance.SGLE)
                        self.lu_system = self.spm.lowlevel.System(self.spm.lowlevel.System.Instance.SGLE)
                        self.connected = True
                    else:
                        self.logger.error("Lowlevel Scripting interface of spm controller is not enabled.")
                else:
                    self.logger.error("Scripting interface of spm controller is not enabled.")
                self.spm_system.MotorCustomSpeedFactor(self.motor_speed)
            else:
                self.logger.error("Cannot connect to spm controller. spm of None was provided")
        return self.connected

    def is_connected(self) -> bool:
        """ return True if motor control is connected to a spm controller"""
        return self.connected
    
    @property
    def motor_speed(self) -> float:
        """ returns the current set motor speed used far all motor movements"""
        return self.__motor_speed

    @motor_speed.setter   
    def motor_speed(self, speed: float):
        """ Set all motor speeds. 
        Parameter
        ----------
            val: float
                speed values range is: 0.01 > value <=1.0. 
                1.0 means full speed 
        """
        self.__motor_speed = min(max(0.01, speed), 1.0)

    @property
    def sleep_time_after_motor_command(self):
        """ Current sleep time in [s] after each COM motor command.  """
        return self.__sleep_time_after_motor_command
    
    @sleep_time_after_motor_command.setter
    def sleep_time_after_motor_command(self, val):
        """ Sleep time in [s] after each COM motor command. 
            It already happened, when two motor move command to separate axes were sent through the COM interface in a row, 
            there was a race condition and the same axis got twice the same command.
            Therefore it is suggested to set this parameter to 200ms, or to zero; when you don't want to wait. 
        """
        self.__sleep_time_after_motor_command = min(max(0.0, val), 1.0)

    @property
    def status_poll_time(self):
        """ Sleep time in [s] during status polling. """
        return self.__status_poll_time
    
    @status_poll_time.setter
    def status_poll_time(self, val):
        """ Sleep time in [s] during status polling.
        """
        self.__status_poll_time = min(max(0.0, val), 1.0)
    
    @property
    def max_wait_time(self):
        """ Maximal time in [s] to wait for end of movement is reached. """
        return self.__max_wait_time
    
    @max_wait_time.setter
    def max_wait_time(self, val):
        """ Maximal time in [s] to wait for end of movement is reached.
        """
        self.__max_wait_time = min(max(0.0, val), 120.0)
        
    def get_motor_status(self, motor_id: MotorID) -> MotorState:
        """ collects current axis states for a motor"""
        status = MotorState()
        self.lu_direct_motor_ctrl.current_axis.value = motor_id.LUIndex
        axis_status = self.lu_direct_motor_ctrl.axis_status.value
        limit_switch_status = self.lu_direct_motor_ctrl.axis_limit_status.value

        status.move_state = self.map_move_status[int(axis_status)]
        status.limit_state = self.map_limit_status[int(limit_switch_status)]
        return status

    def get_motor_position(self, motor_id: MotorID) -> float:
        """ Returns the current position of a motor in [m]"""
        cur_pos = self.spm_system.GetMotorPosition(motor_id.ProxyIndex)
        return cur_pos

    def get_motor_full_range(self, motor_id: MotorID) -> float:
        """ return the full travel range of a specified motor"""
        return motor_id.Range

    def start_move_relative(self, motor_id: typing.Union[MotorID, list[MotorID]], step_size: typing.Union[float, list[float]]):
        """ Start a relative move from current position 
        
        Paramters
        ---------
        motor_id:  single motorID or list of motorIDs
        step_size: float or list of float
            travel length. can be positive or negative

        """
        self.logger.debug("start motor(s) relative move")
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        self._move_command(motor_list, step_size, MotorControl._MoveMode.Relative, wait_for_end=False)

    def do_move_relative(self, motor_id: typing.Union[MotorID, list[MotorID]], step_size: typing.Union[float, list[float]]) -> bool:
        """ start and wait a relative move from current position 
        
        Paramters
        ---------
        motor_id:  single motorID or list of motorIDs
        step_size: float or list of float
            travel length. can be positive or negative

        """
        self.logger.debug("do motor(s) relative move")
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        return self._move_command(motor_list, step_size, MotorControl._MoveMode.Relative, wait_for_end=True)

    def start_move_absolute(self, motor_id: typing.Union[MotorID, list[MotorID]], target_position: typing.Union[float, list[float]]):
        """ Start a move to absolute target position(s) 
        
        Paramters
        ---------
        motor_id:  single motorID or list of motorIDs
        target_position: float or list of float
            travel length. can be positive or negative

        """
        self.logger.debug("start motor(s) move to absolute position")
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        self._move_command(motor_list, target_position, MotorControl._MoveMode.Absolute, wait_for_end=False)

    def do_move_absolute(self, motor_id: typing.Union[MotorID, list[MotorID]], target_position: typing.Union[float, list[float]]) -> bool:
        """ start and wait a move to absolute target position(s) 
        
        Paramters
        ---------
        motor_id:  single motorID or list of motorIDs
        target_position: float or list of float
            target position has to be positive

        """
        self.logger.debug("do motor(s) move to absolute position")
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        return self._move_command(motor_list, target_position, MotorControl._MoveMode.Absolute, wait_for_end=True)

    def start_moving(self, motor_id: typing.Union[MotorID, list[MotorID]], direction: typing.Union[Direction, list[Direction]]):
        """ Start motor(s) movement in one direction, without stopping
        
        Paramters
        ---------
        motor_id:   single motor_id or list of motor_ids 
        direction:  single direction or list of directions
        """
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id

        if isinstance(direction, Direction):
            direction_list = [direction for i in motor_list]
        else:
            direction_list = direction

        target_positions = []
        for dir in direction_list:
            end_pos = 1.0 if dir == Direction.Forward else -1.0
            target_positions.append(end_pos)

        self._move_command(motor_list, target_positions, MotorControl._MoveMode.Relative, wait_for_end=False)        

    def stop_all_motors(self):
        self.logger.debug("Stopped All Motors")
        self.spm.application.System.MotorStop() 
        time.sleep(self.sleep_time_after_motor_command)

    def set_motor_zero_position(self, motor_id: typing.Union[MotorID, list[MotorID]]):
        """ Set the current position of the specific axis to zero
        Paramter:
        ------------------
        motorID:    single MotorID or list of MotorID
        """
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        for index, motor in enumerate(motor_list):
            self.spm_system.MotorSetPosZero(motor.ProxyIndex)
            time.sleep(self.sleep_time_after_motor_command)

    def start_motor_referencing(self, motor_id: typing.Union[MotorID, list[MotorID]]):
        """ Start the referencing movement the specific motor(s) to the lower limit switch and sets the position to zero
        Paramter:
        ------------------
        motorID:    single motorID of class MotorID or array of motorIDs
        """
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        for index, motor in enumerate(motor_list):
            self.lu_direct_motor_ctrl.current_axis.value = motor.LUIndex
            self.lu_direct_motor_ctrl.axis_speed_factor.value = 100
            # reference
            self.lu_direct_motor_ctrl.search_reference()

    def do_motor_referencing(self, motor_id: typing.Union[MotorID, list[MotorID]]) -> bool:
        """References the specific motor(s) to the lower limit switch and sets the position to zero
        Paramter:
        ------------------
        motorID:    single motorID of class MotorID or array of motorIDs
        """
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        self.start_motor_referencing(motor_list)
        done = self._wait_for_end_of_movement(motor_list)
        if done:
            self.logger.info("Motors are referenced and at lower limit")
        else:
            self.logger.error("Move ended with timeout")
        return done
        
    def do_motor_reference_and_center(self, motor_id: typing.Union[MotorID, list[MotorID]]):
        """References the specific motor(s) to the lower limit switch, sets the position to zero and move to halve of the max range
        Paramter:
        ------------------
        motorID:    single motorID of class MotorID or array of motorIDs
        """
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        done = self.do_motor_referencing(motor_list)
        if done:
            target_positions = []
            for index, motor in enumerate(motor_list):
                target_positions.append(motor.Range/2.0)
            done = self.do_move_absolute(motor_list, target_positions)
        return done

    def do_reference_and_move_back(self, motor_id: typing.Union[MotorID, list[MotorID]]):
        """Reference the specific motor(s) to the lower limit switch and move back to the initial position and update the referenced position"""
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        attrMotVector = []
        for index, motor in enumerate(motor_list):
            attrMotVector.append(motor.ProxyIndex)
        self.spm_app.SPMCtrlManager.LogicalUnit.AttributeVectorDouble(
             self.spm.lowlevel.System._type_id, self.spm.lowlevel.System.Instance.SGLE, 
             self.lu_system.motor_selection._id, attrMotVector)
        self.lu_system.motor_reference_and_move_back()
        done = self._wait_for_end_of_movement(motor_list)
        return done        
        
    def get_max_axis_speed(self, motor_id: MotorID) -> float:
        """ Returns the maximal speed of a motor in [?]"""
        self.lu_direct_motor_ctrl.current_axis.value = motor_id.LUIndex
        max_speed = self.lu_direct_motor_ctrl.axis_max_speed.value
        return max_speed

    def overwrite_max_axis_speed(self, motor_id: typing.Union[MotorID, list[MotorID]], new_max_speed: float):
        """ update the max speed of motors defined in the hardware.
        WARNING: 
        --------
        Please never change the maximum motor speed, without safety checks. 
        It can damage the hardware.
        The max speed has no upper limit as it overwrites the software defined dmc max speed
        """
        motor_list = [motor_id] if isinstance(motor_id, MotorID) else motor_id
        for index, motor in enumerate(motor_list):
            self.lu_direct_motor_ctrl.current_axis.value = motor.LUIndex
            self.lu_direct_motor_ctrl.axis_max_speed.value = new_max_speed
        time.sleep(self.sleep_time_after_motor_command)
    
    # ------- implementation ------------------------------------------

    def _move_command(self, motors: list[MotorID], move_positions: typing.Union[float, list[float]], move_mode: _MoveMode, wait_for_end: bool) -> bool:
        """ handle the moving command """
        self.lu_direct_motor_ctrl.reset_transaction()
        for index, motor in enumerate(motors):
            move_pos = move_positions[index] if isinstance(move_positions, list) else move_positions
            relative_mode = self.lu_direct_motor_ctrl.TransactionMoveRelative.Relative if (move_mode == MotorControl._MoveMode.Relative) else self.lu_direct_motor_ctrl.TransactionMoveRelative.Absolute
            
            self.lu_direct_motor_ctrl.current_axis.value = motor.LUIndex
            self.lu_direct_motor_ctrl.axis_speed_factor.value = self._get_lu_motor_speed()
            self.lu_direct_motor_ctrl.transaction_move_relative.value = relative_mode
            self.lu_direct_motor_ctrl.transaction_move_destination.value = move_pos
            self.lu_direct_motor_ctrl.add_transaction_move()

        self.lu_direct_motor_ctrl.commit_transaction()
        time.sleep(self.sleep_time_after_motor_command)

        if wait_for_end:
            return self._wait_for_end_of_movement(motors)
        return True

    def _wait_for_end_of_movement(self,  motors: list[MotorID]) -> bool:
        """ Wait until all motors are stopped or a timeout is reached.
        
        Paramters
        ---------
        motors: list of motorIDs

        Returns
        -------
         done: bool
            return True if all defined motors are stopped in timout time. False if Timeout was reached
        """
        motor_moving = True
        time_out_reached = False
        start_time = time.time()
        already_checked_once = False
        while motor_moving and not time_out_reached and not already_checked_once: 
            stopped = [] # list of stop state for each motor
            for index, motor in enumerate(motors):
                status = self.get_motor_status(motor)
                if (status.move_state == MoveStatus.IdleReferenced) or (status.move_state == MoveStatus.IdleUnreferenced):
                    stopped.append(True) 
                else:
                    stopped.append(False)

            if all(stopped) and already_checked_once:
                motor_moving = False
                self.logger.debug("All motor stopped")
            elif(all(stopped) and not already_checked_once):
                already_checked_once = True # We check once more
                self.logger.debug("Check again whether motors stopped")
            else:
                if (time.time() - start_time) > self.max_wait_time:
                    time_out_reached = True
                    self.logger.debug("Motors where still moving after timeout reached")
            
            self.tick_callback(time.time() - start_time)
            time.sleep(self.status_poll_time)
        return not time_out_reached

    def _get_lu_motor_speed(self) -> float:
        """ for LU it does not use speed factor between 0 and 1 but between 0 and 100"""
        return self.motor_speed * 100.0

    def default_tick_callback(self, time: float):
        pass


# /// Copyright (C) Nanosurf AG - All Rights Reserved (2021)
# /// Unauthorized copying of this file, via any medium is strictly prohibited
# /// https://www.nanosurf.com
# ///

from nanosurf.lib.gui.nsf_sci_edit import NSFSciEdit
from nanosurf.lib.gui.nsf_tables import NSFNameValueTable, NSFTable, TableEntryIDs
from nanosurf.lib.gui.nsf_info_box import NSFInfoBox
from nanosurf.lib.gui.nsf_plots import NSFChart, NSFColormap
from nanosurf.lib.gui.nsf_sci_edit import NSFSciEdit
from nanosurf.lib.gui.nsf_edit import NSFEdit
from nanosurf.lib.gui.nsf_combobox import NSFComboBox, NSFComboEntry
from nanosurf.lib.gui.bind_gui import connect_to_property
import nanosurf.lib.gui.nsf_colors as nsf_colors




"""Package for scripting the Nansurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

import nanosurf.lib.spm.lowlevel.ctrlunits as ctrlunits

class _CtrlUnitChannelMux(ctrlunits._CtrlUnit):
    def __init__(self, spm):
        self._channelmap = None
        self._channel_name = ""

    @property
    def channelmap(self):
        return self._channelmap

    @channelmap.setter
    def channelmap(self, val):
        self._channelmap = val

    @property
    def channel(self):
        return self._channel_name

    @channel.setter
    def channel(self, val):
        self._channel_name = val
        self.target._lu.input.value = self._channelmap[val]

class CtrlUnitDACHiResMux(_CtrlUnitChannelMux):
    def __init__(self, spm):
        super().__init__(spm)
        self.channelmap = spm.lowlevel.AnalogHiResOut.InputChannels

class CtrlUnitDACFastMux(_CtrlUnitChannelMux):
    def __init__(self, spm):
        super().__init__(spm)
        self.channelmap = spm.lowlevel.AnalogFastOut.InputChannels

class CtrlUnitLockInMux(_CtrlUnitChannelMux):
    def __init__(self, spm):
        super().__init__(spm)
        self.channelmap = spm.lowlevel.SignalAnalyzer.Input




""" The base class for functional modules
Copyright Nanosurf AG 2021
License - MIT
"""
import sys
import logging
from PySide2 import QtWidgets, QtCore

control_layout_width = 200

class CtrlSpacer(QtWidgets.QSpacerItem):
    def __init__(self):
        super().__init__(control_layout_width,40, QtWidgets.QSizePolicy.Fixed,QtWidgets.QSizePolicy.MinimumExpanding)

class ModuleBase(QtCore.QObject):
    def __init__(self, app, gui):
        super().__init__()
        self.app = app
        self.ui = gui
        self.settings = None
        self.__name = ""
        self.logger = logging.getLogger(self.__name)
 
    def start(self):
        self.app.load_settings(self)

        self.do_start()
        if self.ui != None:
            self.ui.create_screen(self)

    @property
    def gui(self):
        return self.ui

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name):
        self.__name = name
        self.logger = logging.getLogger(self.__name)

    def stop(self):
        self.do_stop()
        self.app.save_settings(self)

    def do_start(self):
        raise NotImplementedError(f"Subclass of '{self.__class__.__name__}' has to implement '{sys._getframe().f_code.co_name}()'")

    def do_stop(self):
        raise NotImplementedError(f"Subclass of '{self.__class__.__name__}' has to implement '{sys._getframe().f_code.co_name}()'")

class ModuleScreen(QtWidgets.QWidget):
    def __init__(self):
        self.module = None
        super().__init__()

    def create_screen(self, module: ModuleBase):
        """
        lets create the gui. Creates all GUI elements but no action yet. 
        This is done later in ConnectGUI()
        """
        self.module = module
        self.do_setup_screen(module)

    def do_setup_screen(self, module: ModuleBase):
        raise NotImplementedError(f"Subclass of '{self.__class__.__name__}' has to implement '{sys._getframe().f_code.co_name}()'")




# from setuptools import setup
# setup()
import os
from setuptools import setup, find_packages

def load_doc_file(readme_file_path: str) -> str:
    doc_str = ""
    with open(readme_file_path, "r", encoding="utf-8") as fh:
        doc_str = fh.read()
    return doc_str

def load_version(readme_file_path: str) -> str:
    ver_str = ""
    with open(readme_file_path, "r", encoding="utf-8") as fh:
        ver_str = fh.read()
    _, version = ver_str.split("=")
    version = version.replace('"', "").replace(" ", "")
    return version

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths

package_data_files = []
package_data_files += package_files('nanosurf/app')
package_data_files += package_files('nanosurf/doc')
package_data_files += package_files('nanosurf_internal/app')
package_data_files += package_files('nanosurf_internal/doc')

long_description_file = load_doc_file('nanosurf/doc/README.md')

setup(
    name='nanosurf',
    version=load_version('nanosurf/_version.py'),
    author='Nanosurf AG',
    author_email='scripting@nanosurf.com',
    description='Python API for Nanosurf controllers',
    long_description=long_description_file,
    long_description_content_type="text/markdown",
    license='MIT',
    packages=find_packages(
        include=['*'],
    ),
    package_data={'': package_data_files},
    include_package_data = False,
    zip_safe=False,
    install_requires=['pywin32', 'matplotlib', 'numpy', 'scipy', 'notebook', 'pyside2', 'pyqtgraph', 'psutil', 'debugpy', 'lupa'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows"
    ],
    entry_points={
        'console_scripts': [
            'nanosurf_help = nanosurf:help',
        ]
    },
    python_requires='>=3.9'
)




""" Some helpful file and folder functions
Copyright Nanosurf AG 2021
License - MIT
"""
import os
import pathlib
from datetime import datetime

def create_filename_with_timestamp(base_name: str, extension: str = '.dat', separator: str = "_") -> str:
    """Make filename"""
    current_datetime = datetime.now().strftime("%Y%m%d-%H%M%S")
    filename = base_name + separator + current_datetime + extension
    return filename

def create_folder(file_path: pathlib.Path) -> bool:
    """ Make sure the folder exists. 
    If needed, it  creates the intermediate directories starting from root.
    """
    done = False
    try:
        if not file_path.is_dir():
            os.makedirs(file_path, exist_ok=True)
        done = file_path.is_dir()
    except:
        pass
    return done

def create_unique_folder(base_name: str, folder: pathlib.Path, add_timestamp: bool = True, separator: str = '_') -> pathlib.Path:
    """Create a unique folder.  The folder name has the structure of 'base_name_timestamp_index'
    The timestamp is optional, set argument add_timestamp=False to suppress it.
    A index is added, if the base_name folder (with optional timestamp) is not unique allready. 
    The index starts at zero and is incremented until a unique name is found.

    Parameter
    ---------
    base_name: str
        The new folder name has the structure of 'base_name_timestamp_index'
    folder: pathlib.Path
        The folder in which the new subfolder with the unique name shall be created
    add_timestamp: bool, optional, defaults to True
        If set to True, a timestamp in the format of %Y%m%d-%H%M%S is added
    separator : str, optional, defaults to '_'
        The separation character use to separate parts of the folder name. 
    """
    if add_timestamp:
        base_name = create_filename_with_timestamp(base_name, extension="", separator=separator)

    done = create_folder(folder)
    if done:
        data_folder_name = base_name
        filepath = pathlib.Path(folder) / pathlib.Path(data_folder_name)
        i = 0
        while filepath.is_dir():
            data_folder_name = f"{base_name}_{i:03d}"
            filepath = pathlib.Path(folder) / pathlib.Path(data_folder_name)
            i += 1
        done = create_folder(filepath)
    if not done:
        filepath = pathlib.Path("")
    return filepath




""" The functional module where the functionality goes
Copyright Nanosurf AG 2021
License - MIT
"""
import numpy as np
import pathlib
from PySide2.QtCore import Signal
import nanosurf as nsf
from app import module_base, app_common
from app.app_base import ApplicationBase
from modules.switching_spec_module import settings, worker_task

class SwitchingSpecModule(module_base.ModuleBase):

    sig_work_start_requested = Signal()
    sig_work_stop_requested = Signal()
    sig_work_active = Signal()
    sig_work_done = Signal()
    sig_new_data_available = Signal()
    sig_data_invalid = Signal()

    """ Initialization functions of the module """

    def __init__(self, app: ApplicationBase, gui):
        super().__init__(app, gui)
        self.app = app
        """ Prepare here module settings which are stored and loaded from file by the app framework """
        self.settings = settings.SpecSettings()
        self.result = settings.SpecResults()

    def do_start(self):
        """ This function is called once at startup of application
            Initialize here all module specific values.
        """
        self.setup_worker_thread()
        self.connect_to_properties()

    def do_stop(self):
        """ This function is called at module shutdown"""
        if self.worker_thread.is_thread_running():
            self.logger.info("Wait until worker thread has ended...")
            self.worker_thread.stop_thread(wait=True)

    def connect_to_properties(self):
        """ Connect action functions to settings 
            The connected functions are called whenever a setting is changed (e.g. by GUI elements)
        """
        self.settings.output_span.sig_value_changed.connect(self.update_worker_parameter)
        self.settings.output_center.sig_value_changed.connect(self.update_worker_parameter)
        self.settings.number_of_steps.sig_value_changed.connect(self.update_worker_parameter)
        self.settings.time_delay_after_step.sig_value_changed.connect(self.update_worker_parameter)
        self.settings.output_id.sig_value_changed.connect(self.update_worker_parameter)

    def setup_worker_thread(self):
        """ Create the background worker task and connect to its event """
        self.worker_thread = worker_task.SpecWorker(self)
        self.worker_thread.sig_worker_started.connect(self._on_sig_worker_started)
        self.worker_thread.sig_worker_finished.connect(self._on_sig_worker_finished)
        self.worker_thread.sig_new_data.connect(self._on_sig_worker_new_data)
        self.worker_thread.sig_tick.connect(self._on_sig_worker_tick)
        self.worker_thread.start_thread()

    """ Now the business logic of the module """
        
    def start_worker(self):
        if not self.worker_thread.is_worker_running():
            self.result = settings.SpecResults()
            self.update_worker_parameter()
            self.sig_work_start_requested.emit()
            self.worker_thread.start_worker()

    def stop_worker(self):
        if self.worker_thread.is_worker_running():
            self.sig_work_stop_requested.emit()
            self.worker_thread.abort_worker()
    
    def is_worker_busy(self) -> bool:
        return self.worker_thread.is_worker_running()

    def get_result(self) -> settings.SpecResults:
        return self.result

    def get_worker_result(self) -> worker_task.SpecWorkerData:
        return self.worker_thread.get_result()

    def update_worker_parameter(self):
        self.worker_thread.par_output_id = self.settings.output_id.value
        self.worker_thread.par_output_span = self.settings.output_span.value
        self.worker_thread.par_output_center = self.settings.output_center.value
        self.worker_thread.par_time_per_step = self.settings.time_delay_after_step.value
        self.worker_thread.par_steps = self.settings.number_of_steps.value
        self.sig_data_invalid.emit()

    """ worker thread state handling """

    def _on_sig_worker_started(self):
        self.app.show_message("Working ...") 
        self.logger.info("Thread started to work")
        self.sig_data_invalid.emit()
        self.sig_work_active.emit()

    def _on_sig_worker_finished(self):
        self.sig_work_done.emit()
        if not self.worker_thread.is_worker_aborted():
            self.app.show_message("Work done") 
            self.logger.info("Background worker finished ")
            if self.settings.auto_save_data.value:
                if len(self.get_result().outputs) > 0:
                    self.save_data()
                else:
                    self.app.show_message("Auto save info: No data measured") 
                    self.logger.info("Auto save info: No data measured")
        else:
            self.app.show_message("Work aborted") 
            self.logger.info("Background worker aborted ")
 
    def _on_sig_worker_tick(self):
        self.app.show_message("Tick") 

    def _on_sig_worker_new_data(self):
        data = self.worker_thread.get_result()
        self.result.outputs = np.append(self.result.outputs, data.output)
        self.result.outputs_unit = data.output_unit
        self.result.amplitudes_unit = data.amplitude_unit
        print(data.amplitude_off)
        self.result.amplitudes_off = np.append(self.result.amplitudes_off, data.amplitude_off)
        self.result.amplitudes_on = np.append(self.result.amplitudes_on, data.amplitude_on)
        self.result.phases_off = np.append(self.result.phases_off, data.phase_off)
        self.result.phases_on = np.append(self.result.phases_on, data.phase_on)
        self.sig_new_data_available.emit()

    def save_data(self):
        folder = self.settings.folder_name.value
        basename = self.settings.file_name_mask.value
        if nsf.util.fileutil.create_folder(folder):
            current_data = self.get_result()
            dataset = {}
            dataset[0] = current_data.outputs
            dataset[1] = current_data.amplitudes_on
            dataset[2] = current_data.amplitudes_off
            dataset[3] = current_data.phases_on
            dataset[4] = current_data.phases_off

            file_name = f"{basename}_{int(self.settings.file_index.value):03d}.csv"
            file_path = pathlib.Path(folder) / pathlib.Path(file_name)
            nsf.util.dataexport.savedata_txt(file_path, dataset, header="Voltage[V]; Amp_on[V]; Amp_off[V]; Phase_on[°]; Phase_off[°]", separator=";")
           
            self.settings.file_index.value += 1

            self.app.show_message(f"Data saved to: {str(file_path)}") 
        else:
            self.app.show_message(f"Error: Could not save data to folder: {str(folder)}", app_common.MsgType.Error) 




"""Package for scripting the Nansurf control software.
Copyright (C) Nanosurf AG - All Rights Reserved (2021)
License - MIT"""

from nanosurf.lib.spm.lowlevel import ctrlunits
import nanosurf.lib.spm.lowlevel.ctrlunits.dac as dac
import nanosurf.lib.spm.lowlevel.ctrlunits.adc as adc
import nanosurf.lib.spm.lowlevel.ctrlunits.analyzer as analyzer
import nanosurf.lib.spm.lowlevel.ctrlunits.channelmux as channelmux
import nanosurf.lib.spm.lowlevel.ctrlunits.capture as capture
import nanosurf.lib.spm.lowlevel.ctrlunits.sampler as sampler

class _CtrlUnitFactory:
    """Main class for working with the CX """

    def __init__(self, spm: 'Spm' = None):
        """
        Parameters
        ----------
        spm
            reference to the nanosurf.spm class used
        """
        self._spm = spm
        self._dacmap = {
            ctrlunits.DAC.HiResOut_POSITIONX:["AnalogHiResOut", "POSITIONX"],
            ctrlunits.DAC.HiResOut_POSITIONY:["AnalogHiResOut", "POSITIONY"],
            ctrlunits.DAC.HiResOut_POSITIONZ:["AnalogHiResOut", "POSITIONZ"],
            ctrlunits.DAC.HiResOut_POSITIONW:["AnalogHiResOut", "POSITIONW"],
            ctrlunits.DAC.HiResOut_TIPVOLTAGE:["AnalogHiResOut", "TIPVOLTAGE"],
            ctrlunits.DAC.HiResOut_OUT6:["AnalogHiResOut", "APPROACH"],
            ctrlunits.DAC.HiResOut_OUT7:["AnalogHiResOut", "OUT7"],
            ctrlunits.DAC.HiResOut_OUT8:["AnalogHiResOut", "OUT8"],
            ctrlunits.DAC.HiResOut_USER1:["AnalogHiResOut", "USER1"],
            ctrlunits.DAC.HiResOut_USER2:["AnalogHiResOut", "USER2"],
            ctrlunits.DAC.HiResOut_USER3:["AnalogHiResOut", "USER3"],
            ctrlunits.DAC.HiResOut_USER4:["AnalogHiResOut", "USER4"],
            ctrlunits.DAC.FastOut_EXCITATION:["AnalogFastOut", "EXCITATION"],
            ctrlunits.DAC.FastOut_USER:["AnalogFastOut", "USER"],
            ctrlunits.DAC.FastOut_FAST2:["AnalogFastOut", "FAST2"],
            ctrlunits.DAC.FastOut_FAST3:["AnalogFastOut", "FAST3"],
        }
        self._adcmap = {
            ctrlunits.ADC.HiResIn_DEFLECTION:["AnalogHiResIn", "DEFLECTION"],
            ctrlunits.ADC.HiResIn_LATERAL:["AnalogHiResIn", "LATERAL"],
            ctrlunits.ADC.HiResIn_POSITIONX:["AnalogHiResIn", "POSITIONX"],
            ctrlunits.ADC.HiResIn_POSITIONY:["AnalogHiResIn", "POSITIONY"],
            ctrlunits.ADC.HiResIn_POSITIONZ:["AnalogHiResIn", "POSITIONZ"],
            ctrlunits.ADC.HiResIn_DETECTORSUM:["AnalogHiResIn", "DETECTORSUM"],
            ctrlunits.ADC.HiResIn_TIPCURRENT:["AnalogHiResIn", "TIPCURRENT"],
            ctrlunits.ADC.HiResIn_IN6:["AnalogHiResIn", "IN6"],
            ctrlunits.ADC.HiResIn_USER1:["AnalogHiResIn", "USER1"],
            ctrlunits.ADC.HiResIn_USER2:["AnalogHiResIn", "USER2"],
            ctrlunits.ADC.HiResIn_USER3:["AnalogHiResIn", "USER3"],
            ctrlunits.ADC.HiResIn_USER4:["AnalogHiResIn", "USER4"],
            ctrlunits.ADC.FastIn_DEFLECTION:["AnalogFastIn", "DEFLECTION"],
            ctrlunits.ADC.FastIn_CH2:["AnalogFastIn", "CH2"],
            ctrlunits.ADC.FastIn_USER :["AnalogFastIn", "USER"],
        }
        self._analyzermap = {
            ctrlunits.Analyzer.SigAnalyzer_1:["SignalAnalyzer", "INST1"],
            ctrlunits.Analyzer.SigAnalyzer_2:["SignalAnalyzer", "INST2"],
        }

    def get_dac(self, ctrlunit_id: ctrlunits.DAC) -> dac._CtrlUnitDAC:
        if self._dacmap[ctrlunit_id][0] == "AnalogHiResOut":
            return dac.CtrlUnitDACHires(self._spm, self._dacmap[ctrlunit_id][1])
        elif self._dacmap[ctrlunit_id][0] == "AnalogFastOut":
            return dac.CtrlUnitDACFast(self._spm, self._dacmap[ctrlunit_id][1])
        return None

    def get_adc(self, ctrlunit_id: ctrlunits.ADC) -> adc._CtrlUnitADC:
        if self._adcmap[ctrlunit_id][0] == "AnalogHiResIn":
            return adc.CtrlUnitADCHiRes(self._spm, self._adcmap[ctrlunit_id][1])
        elif self._adcmap[ctrlunit_id][0] == "AnalogFastIn":
            return adc.CtrlUnitADCFast(self._spm, self._adcmap[ctrlunit_id][1])
        return None

    def get_sin_generator(self, ctrlunit_id: ctrlunits.Analyzer = ctrlunits.Analyzer.SigAnalyzer_2) -> analyzer.CtrlUnitsSineWaveGenerator:
        if self._analyzermap[ctrlunit_id][0] == "SignalAnalyzer":
            return analyzer.CtrlUnitsSineWaveGenerator(self._spm, self._analyzermap[ctrlunit_id][1])
        return None

    def get_lock_in(self, ctrlunit_id: ctrlunits.Analyzer = ctrlunits.Analyzer.SigAnalyzer_2) -> analyzer.CtrlUnitLockIn:
        if self._analyzermap[ctrlunit_id][0] == "SignalAnalyzer":
            return analyzer.CtrlUnitLockIn(self._spm, self._analyzermap[ctrlunit_id][1])
        return None

    def get_multiplexer(self, ctrlunit_id: ctrlunits.ChannelMux) -> channelmux._CtrlUnitChannelMux:
        if ctrlunit_id == ctrlunits.ChannelMux.AnalogHiResOut:
            return channelmux.CtrlUnitDACHiResMux(self._spm)
        elif ctrlunit_id == ctrlunits.ChannelMux.AnalogFastOut:
            return channelmux.CtrlUnitDACFastMux(self._spm)
        elif ctrlunit_id == ctrlunits.ChannelMux.SigAnalyzer_1:
            return channelmux.CtrlUnitLockInMux(self._spm)
        elif ctrlunit_id == ctrlunits.ChannelMux.SigAnalyzer_2:
            return channelmux.CtrlUnitLockInMux(self._spm)
        return None

    def get_data_capture(self, ctrlunit_id: ctrlunits.DataCapture) -> capture._CtrlUnitCapture:
        if ctrlunit_id == ctrlunits.DataCapture.CaptureHiRes:
            return capture.CtrlUnitCaptureHiRes(self._spm)
        elif ctrlunit_id == ctrlunits.DataCapture.CaptureFast:
            return capture.CtrlUnitCaptureFast(self._spm)
        return None

    def get_data_sampler(self, sampler_mod = ctrlunits.DataSampler.SamplerHiRes) -> sampler._CtrlUnitSampler:
        if sampler_mod == ctrlunits.DataSampler.SamplerHiRes:
            return sampler.CtrlUnitSamplerHiRes(self._spm)
        return None